# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import sys

import numpy as np

from karabo.common.api import (
    KARABO_SCHEMA_DISPLAY_TYPE, KARABO_SCHEMA_MAX_EXC, KARABO_SCHEMA_MAX_INC,
    KARABO_SCHEMA_MAX_SIZE, KARABO_SCHEMA_MIN_EXC, KARABO_SCHEMA_MIN_INC,
    KARABO_SCHEMA_MIN_SIZE, KARABO_SCHEMA_VALUE_TYPE)
from karabo.native import Hash, Schema
from karabogui.binding.api import (
    BoolBinding, FloatBinding, Int8Binding, Int16Binding, Int32Binding,
    Int64Binding, StringBinding, Uint8Binding, Uint16Binding, Uint32Binding,
    Uint64Binding, VectorBoolBinding, VectorDoubleBinding, VectorFloatBinding,
    VectorHashBinding, VectorInt8Binding, VectorInt16Binding,
    VectorInt32Binding, VectorInt64Binding, VectorStringBinding,
    VectorUint8Binding, VectorUint16Binding, VectorUint32Binding,
    VectorUint64Binding, attr_fast_deepcopy, get_dtype_format, get_min_max,
    get_min_max_size, get_native_min_max, has_min_max_attributes,
    is_signed_vector_integer, is_unsigned_vector_integer, is_vector_integer,
    realign_hash)


def test_simple_int_min_max():
    assert get_min_max(Int8Binding()) == (-(1 << 7), (1 << 7) - 1)
    assert get_min_max(Int16Binding()) == (-(1 << 15), (1 << 15) - 1)
    assert get_min_max(Int32Binding()) == (-(1 << 31), (1 << 31) - 1)
    assert get_min_max(Int64Binding()) == (-(1 << 63), (1 << 63) - 1)

    assert get_min_max(Uint8Binding()) == (0, (1 << 8) - 1)
    assert get_min_max(Uint16Binding()) == (0, (1 << 16) - 1)
    assert get_min_max(Uint32Binding()) == (0, (1 << 32) - 1)
    assert get_min_max(Uint64Binding()) == (0, (1 << 64) - 1)


def test_native_int_min_max():
    assert get_native_min_max(Int8Binding()) == (-(1 << 7), (1 << 7) - 1)
    assert get_native_min_max(Int16Binding()) == (-(1 << 15), (1 << 15) - 1)
    assert get_native_min_max(Int32Binding()) == (-(1 << 31), (1 << 31) - 1)
    assert get_native_min_max(Int64Binding()) == (-(1 << 63), (1 << 63) - 1)

    assert get_native_min_max(Uint8Binding()) == (0, (1 << 8) - 1)
    assert get_native_min_max(Uint16Binding()) == (0, (1 << 16) - 1)
    assert get_native_min_max(Uint32Binding()) == (0, (1 << 32) - 1)
    assert get_native_min_max(Uint64Binding()) == (0, (1 << 64) - 1)

    # Binding specifications are neglected
    binding = Int8Binding(attributes={
        KARABO_SCHEMA_VALUE_TYPE: 'INT8',
        KARABO_SCHEMA_MIN_EXC: -7,
        KARABO_SCHEMA_MAX_EXC: 43})
    assert get_native_min_max(binding) == (-(1 << 7), (1 << 7) - 1)


def test_native_vector_min_max():
    test_map = [
        (VectorInt8Binding, (-(1 << 7), (1 << 7) - 1)),
        (VectorInt16Binding, (-(1 << 15), (1 << 15) - 1)),
        (VectorInt32Binding, (-(1 << 31), (1 << 31) - 1)),
        (VectorInt64Binding, (-(1 << 63), (1 << 63) - 1)),

        (VectorUint8Binding, (0, (1 << 8) - 1)),
        (VectorUint16Binding, (0, (1 << 16) - 1)),
        (VectorUint32Binding, (0, (1 << 32) - 1)),
        (VectorUint64Binding, (0, (1 << 64) - 1))]
    for binding, expected in test_map:
        assert get_native_min_max(binding()) == expected


def test_ranged_int_min_max():
    binding = Int8Binding(attributes={KARABO_SCHEMA_MIN_EXC: -7,
                                      KARABO_SCHEMA_MAX_EXC: 43})
    assert get_min_max(binding) == (-6, 42)

    binding = Int8Binding(attributes={KARABO_SCHEMA_MIN_INC: -7,
                                      KARABO_SCHEMA_MAX_INC: 43})
    assert get_min_max(binding) == (-7, 43)


def test_simple_float_min_max():
    binding = FloatBinding(attributes={KARABO_SCHEMA_VALUE_TYPE: 'DOUBLE'})
    assert get_min_max(binding) == (-sys.float_info.max, sys.float_info.max)
    assert get_native_min_max(binding) == (-sys.float_info.max,
                                           sys.float_info.max)


def test_unsupported():
    assert get_min_max(BoolBinding()) == (None, None)
    assert get_native_min_max(BoolBinding()) == (None, None)


def test_vector_min_max():
    binding = VectorInt32Binding(attributes={KARABO_SCHEMA_MIN_SIZE: 1,
                                             KARABO_SCHEMA_MAX_SIZE: 2})
    assert get_min_max_size(binding) == (1, 2)

    binding = VectorBoolBinding(attributes={KARABO_SCHEMA_MIN_SIZE: 43})
    assert get_min_max_size(binding) == (43, None)

    binding = VectorFloatBinding(attributes={KARABO_SCHEMA_MAX_SIZE: 17})
    assert get_min_max_size(binding) == (None, 17)


def test_has_min_max():
    binding = FloatBinding(attributes={KARABO_SCHEMA_MIN_INC: 1,
                                       KARABO_SCHEMA_MAX_EXC: 3})
    assert has_min_max_attributes(binding) is True

    binding = FloatBinding(attributes={KARABO_SCHEMA_MIN_EXC: 1,
                                       KARABO_SCHEMA_MAX_EXC: 3})
    assert has_min_max_attributes(binding) is True

    binding = FloatBinding(attributes={KARABO_SCHEMA_MIN_INC: 1,
                                       KARABO_SCHEMA_MAX_INC: 3})
    assert has_min_max_attributes(binding) is True

    binding = FloatBinding(attributes={KARABO_SCHEMA_MIN_EXC: 1,
                                       KARABO_SCHEMA_MAX_INC: 3})
    assert has_min_max_attributes(binding) is True

    binding = FloatBinding(attributes={KARABO_SCHEMA_MAX_INC: 3})
    assert has_min_max_attributes(binding) is False

    binding = FloatBinding(attributes={KARABO_SCHEMA_MIN_EXC: 1})
    assert has_min_max_attributes(binding) is False

    binding = FloatBinding(attributes={KARABO_SCHEMA_MIN_EXC: 1,
                                       KARABO_SCHEMA_MIN_INC: 5})
    assert has_min_max_attributes(binding) is False

    binding = FloatBinding(attributes={KARABO_SCHEMA_MAX_EXC: 1,
                                       KARABO_SCHEMA_MAX_INC: 5})
    assert has_min_max_attributes(binding) is False


def test_attr_fast_deepcopy():
    # Note: This is as well in MDL conf. Find proper place!
    def _safe_compare(a, b):
        # Use repr() to get around the lack of Schema comparison
        return len(a) == len(b) and all(repr(a[k]) == repr(b[k]) for k in a)

    d = {
        'a': [1, 2, 3],
        'b': {'sub': 42},
        'c': 'Hi there!',
        'd': np.zeros((10,)),
        'e': (1, 2, 3),
        'f': Hash('simple', 32),
        'g': Schema()
    }
    copy = attr_fast_deepcopy(d)
    assert _safe_compare(copy, d)


def test_realign_hash():
    """Test the realign hash function"""

    # 1. Check when ordering was not been changed
    foo_hash = Hash("stringProperty", "foo",
                    "uintProperty", 1,
                    "boolProperty", 2)
    reference = foo_hash.keys()

    assert realign_hash(foo_hash, reference=reference) == foo_hash

    # 2. Check when the hash was jumbled and would still respect the keys order
    jumbled_hash = Hash("uintProperty", 1,
                        "stringProperty", "foo",
                        "boolProperty", 2)
    assert realign_hash(jumbled_hash, reference=reference) == foo_hash

    # 3. Check if the hash contains a new property with respect to reference
    # It must be appended appear at the end of the hash
    new_hash = Hash("stringProperty", "foo",
                    "uintProperty", 1,
                    "intProperty", 1,  # new property
                    "boolProperty", 2)

    reordered_hash = realign_hash(new_hash, reference=reference)
    copy_hash = Hash(foo_hash)
    copy_hash["intProperty"] = 1  # append the new property at the end
    assert reordered_hash == copy_hash

    # 4. Check if the hash has a deleted property from the keys
    # In the current case, `uintProperty` is removed
    new_hash = Hash("stringProperty", "foo",
                    "boolProperty", 2)
    reordered_hash = realign_hash(new_hash, reference=reference)
    # The removed property gets value of `None` but the Hash still respects
    # the reference key order
    copy_hash = Hash(foo_hash)
    copy_hash["uintProperty"] = None
    assert reordered_hash == copy_hash


def test_display_format():
    assert get_dtype_format(None) == "{}"

    binding = FloatBinding()
    assert get_dtype_format(binding) == "{:.8g}"

    binding = Int8Binding(attributes={KARABO_SCHEMA_DISPLAY_TYPE: "bin"})
    assert get_dtype_format(binding) == "0b{:b}"

    binding = Int8Binding(attributes={KARABO_SCHEMA_DISPLAY_TYPE: "hex"})
    assert get_dtype_format(binding) == "0x{:X}"

    binding = Int8Binding(attributes={KARABO_SCHEMA_DISPLAY_TYPE: "oct"})
    assert get_dtype_format(binding) == "o{:o}"

    binding = Int8Binding(attributes={KARABO_SCHEMA_DISPLAY_TYPE: "oct|||"})
    assert get_dtype_format(binding) == "o{:o}"

    binding = Int8Binding(attributes={KARABO_SCHEMA_DISPLAY_TYPE: "moct"})
    assert get_dtype_format(binding) == "{}"

    # Only accounted for IntegerBindings
    binding = StringBinding(attributes={KARABO_SCHEMA_DISPLAY_TYPE: "oct"})
    assert get_dtype_format(binding) == "{}"


def test_integer_binding():
    # Signed bindings
    assert is_signed_vector_integer(VectorInt8Binding())
    assert is_signed_vector_integer(VectorInt16Binding())
    assert is_signed_vector_integer(VectorInt32Binding())
    assert is_signed_vector_integer(VectorInt64Binding())

    # Unsigned bindings
    assert is_unsigned_vector_integer(VectorUint8Binding())
    assert is_unsigned_vector_integer(VectorUint16Binding())
    assert is_unsigned_vector_integer(VectorUint32Binding())
    assert is_unsigned_vector_integer(VectorUint64Binding())

    # All int bindings
    assert is_vector_integer(VectorInt8Binding())
    assert is_vector_integer(VectorInt16Binding())
    assert is_vector_integer(VectorInt32Binding())
    assert is_vector_integer(VectorInt64Binding())
    assert is_vector_integer(VectorUint8Binding())
    assert is_vector_integer(VectorUint16Binding())
    assert is_vector_integer(VectorUint32Binding())
    assert is_vector_integer(VectorUint64Binding())

    assert not is_vector_integer(VectorBoolBinding())
    assert not is_vector_integer(VectorStringBinding())
    assert not is_vector_integer(VectorHashBinding())
    assert not is_vector_integer(VectorDoubleBinding())



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from collections import namedtuple

import pytest

from karabo.common.scenemodel.api import SceneModel, SceneTargetWindow
from karabogui.api import get_panel_wrangler
from karabogui.events import KaraboEvent
from karabogui.programs.cinema import create_cinema
from karabogui.singletons.network import Network
from karabogui.testing import (
    set_test_organization_info, singletons, system_hash)
from karabogui.topology.api import SystemTopology

namespace = namedtuple("namespace", "host port domain nosplash "
                                    "scene_uuid timeout")


def test_cinema_normal(gui_app, mocker):
    """Test the karabo cinema with a direct connect"""
    ns = namespace("myhost", "myport", "CTRL", True, ["uuid-1231231321"], 1)

    network = Network()
    network.connectToServerDirectly = mocker.Mock()
    network.connectToServerDirectly.return_value = True
    network.onSubscribeLogs = mocker.Mock()
    info = mocker.patch("karabogui.programs.base.set_app_info")
    topology = SystemTopology()
    path = "karabogui.programs.cinema.broadcast_event"
    with singletons(network=network, topology=topology):
        broadcast = mocker.patch(path)
        success, app = create_cinema(ns)
        info.assert_called_with(app, "XFEL", "xfel.eu", "KaraboGUI")
        # Make sure no settings are erased in tests
        set_test_organization_info(app)
        assert success
        network.signalServerConnectionChanged.emit(True)

        network.onSubscribeLogs.assert_not_called()

        # Initialize the trees
        topology.initialize(system_hash())
        broadcast.assert_called_with(
            KaraboEvent.OpenSceneLink,
            {'name': 'CTRL-uuid-1231231321',
             'target_window': SceneTargetWindow.MainWindow,
             'target': 'uuid-1231231321'})


@pytest.mark.parametrize("authenticated,read_only", [
    (True, False), (True, True), (False, False), (False, True), (True, None)])
def test_user_session_button(gui_app, mocker, authenticated, read_only):
    wrangler = get_panel_wrangler()
    target_window = SceneTargetWindow.Dialog
    model = SceneModel(simple_name="Test_scene")

    krb_access = mocker.patch("karabogui.singletons.panel_wrangler.krb_access")
    is_authenticated = mocker.patch(
        "karabogui.singletons.panel_wrangler.is_authenticated")
    is_authenticated.return_value = authenticated
    krb_access.SERVER_READ_ONLY = read_only
    wrangler._open_scene(model, target_window, attached=False)

    panel = wrangler._unattached_scene_panels.get(model)
    assert panel is not None

    expected = authenticated and read_only is False

    assert panel.tbUserSession.isVisible() == expected



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtWidgets import QAction, QMenu
from traits.api import Bool, Dict, Instance, Property, Type, on_trait_change

from ..base.controller import BaseController, ControllerAggregator
from .analyzer import HistogramAnalyzer
from .plot import HistogramPlot
from .stats import HistogramStats


class HistogramController(BaseController):

    # The analyzer class for the histogram plot
    analyzer = Instance(HistogramAnalyzer, args=())

    # The plot item controller class for the histogram plot
    plot_klass = Type(HistogramPlot)


class HistogramAggregator(ControllerAggregator):

    # The histogram aux plot only contains one histogram plot
    contents = Dict({"top": HistogramController})

    # --- Plot configuration ---
    # The stats are shown by default
    show_stats = Bool(True)

    # --- Convenience properties to get/set values on the histogram plot ---
    controller = Property
    colormap = Property
    levels = Property

    def __init__(self, **traits):
        super().__init__(**traits)

        # Create menu
        menu = QMenu()
        show_stats_action = QAction("Show statistics", menu)
        show_stats_action.setCheckable(True)
        show_stats_action.setChecked(self.show_stats)
        show_stats_action.triggered.connect(lambda x:
                                            self.trait_set(show_stats=x))
        menu.addAction(show_stats_action)
        self.controller.plot.menu = menu

    # -----------------------------------------------------------------------
    # Public methods

    def process(self, region):
        self.controller.process(region, axis=0)
        if self.show_stats:
            self.stats = HistogramStats(stats=self.controller.analyzer.stats)

    # -----------------------------------------------------------------------
    # Trait handlers

    @on_trait_change("show_stats")
    def _show_stats(self, show_stats):
        stats = None
        if show_stats:
            stats = HistogramStats(stats=self.controller.analyzer.stats)
        self.stats = stats

    def destroy(self):
        super().destroy()
        self.on_trait_change(self._show_stats, "show_stats", remove=True)

    # -----------------------------------------------------------------------
    # Trait properties

    def _get_controller(self):
        return self.controllers["top"]

    def _set_colormap(self, colormap):
        self.controller.plot.colormap = colormap

    def _set_levels(self, levels):
        self.controller.plot.levels = levels



#############################################################################
# Author: <john.wiggins@xfel.eu>
# Created on August 4, 2017
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy.QtCore import QEvent, QRect, Qt
from qtpy.QtWidgets import (
    QApplication, QStyle, QStyledItemDelegate, QStyleOptionButton)

import karabogui.access as krb_access
from karabo.native import AccessLevel
from karabogui import icons
from karabogui.binding.api import DeviceProxy, SlotBinding

from .utils import (
    ButtonState, get_device_state_string, handle_default_state, set_fill_rect)

ICON_SIZE = 32
ICON_PADDING = 3
__slot_pixmap = None


def _get_button_rect(rect):
    """Adjust a Qt-provided rectangle to the appropriate bounds

    XXX: Revisit this later. It's not clear what the appropriate way to get
    this information is. `QStyle` has methods which should help, but
    QStyle.getSubelementRect(QStyle.SE_ItemViewItemDecoration, ...) doesn't
    quite return the correct rectangle for a row's icon. Not sure why...
    """
    slot_pixmap = _slot_pixmap()
    rect.setLeft(rect.x() + slot_pixmap.size().width() + ICON_PADDING * 2)
    return rect


def _slot_pixmap():
    """Lazily load the slot icon"""
    global __slot_pixmap
    if __slot_pixmap is None:
        __slot_pixmap = icons.slot.pixmap(ICON_SIZE)
    return __slot_pixmap


class SlotButtonDelegate(QStyledItemDelegate):
    """A QStyledItemDelegate for slot buttons
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self._button_states = {}
        self._global_access = krb_access.GLOBAL_ACCESS_LEVEL

    def paint(self, painter, option, index):
        """Reimplemented function of QStyledItemDelegate."""
        model = index.model()
        proxy = model.index_ref(index)
        if not isinstance(getattr(proxy, 'binding', None), SlotBinding):
            super().paint(painter, option, index)
            return

        set_fill_rect(painter, option, index)

        self._draw_icon(painter, option)
        self._draw_button(painter, option, index, proxy)

    def editorEvent(self, event, model, option, index):
        """Reimplemented function of QStyledItemDelegate."""
        handled_types = (QEvent.MouseButtonPress, QEvent.MouseButtonRelease)
        if event.type() in handled_types:
            proxy = model.index_ref(index)
            if isinstance(getattr(proxy, 'binding', None), SlotBinding):
                self._handle_event_state(proxy, event, option)
                return True

        return super().editorEvent(
            event, model, option, index)

    def setAccessLevel(self, level: AccessLevel):
        self._global_access = level

    def _draw_button(self, painter, option, index, proxy):
        """Draw a button for this slot delegate"""
        key = proxy.key
        state = self._button_states.get(key, ButtonState.DISABLED)
        allowed = self.is_allowed(proxy)
        button_state = handle_default_state(allowed, state)
        self._button_states[key] = state
        button = QStyleOptionButton()
        button.state = button_state
        button.rect = _get_button_rect(option.rect)
        button.text = index.data()
        button.features = QStyleOptionButton.AutoDefaultButton
        QApplication.style().drawControl(QStyle.CE_PushButton, button, painter)

    def _draw_icon(self, painter, option):
        """Draw a slot icon"""
        rect = option.rect
        slot_pixmap = _slot_pixmap()
        pix_size = slot_pixmap.size()
        w, h = pix_size.width(), pix_size.height()
        pix_rect = QRect(rect.x() + ICON_PADDING, rect.center().y() - h // 2,
                         w, h)
        QApplication.style().drawItemPixmap(painter, pix_rect, Qt.AlignCenter,
                                            slot_pixmap)

    def is_allowed(self, proxy):
        is_device = isinstance(proxy.root_proxy, DeviceProxy)
        allowed = (is_device and proxy.binding.is_allowed(
            get_device_state_string(proxy.root_proxy))
                   and (self._global_access >=
                        proxy.binding.requiredAccessLevel))
        return allowed

    def _handle_event_state(self, proxy, event, option):
        """Determine the state of a given proxy's button during user
        interaction.
        """
        key = proxy.key
        state = self._button_states.get(key, ButtonState.DISABLED)
        if self.is_allowed(proxy):
            rect = _get_button_rect(option.rect)
            if rect.contains(event.pos()):
                if event.type() == QEvent.MouseButtonPress:
                    state = ButtonState.PRESSED
                elif state == ButtonState.PRESSED:
                    proxy.execute()
            if (state == ButtonState.PRESSED and
                    event.type() == QEvent.MouseButtonRelease):
                state = ButtonState.ENABLED
        self._button_states[key] = state
        return state



# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.

from qtpy.QtCore import Qt

from karabo.common.project.api import MacroModel
from karabogui.project.dialog.move_handle import MoveHandleDialog
from karabogui.testing import click_button


def test_move_handle_dialog(gui_app):
    models = [MacroModel(simple_name="Macro1"),
              MacroModel(simple_name="Macro2")]

    dialog = MoveHandleDialog(models)
    assert len(dialog.items) == 2

    item = dialog.widget.item(0)
    assert item.data(Qt.UserRole).simple_name == "Macro1"

    # move item back and forth
    dialog.widget.setCurrentRow(1)
    click_button(dialog.move_up_button)
    item = dialog.widget.item(0)
    assert item.data(Qt.UserRole).simple_name == "Macro2"
    click_button(dialog.move_down_button)
    item = dialog.widget.item(0)
    assert item.data(Qt.UserRole).simple_name == "Macro1"
    dialog.done(1)
    dialog.done(0)
    dialog.close()
    assert not dialog.isModal()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import operator

from qtpy.QtGui import QPixmap

from karabo.common.scenemodel.api import (
    DigitIconsModel, IconData, SelectionIconsModel, TextIconsModel)
from karabo.native import Configurable, Int32, String
from karabogui.testing import get_class_property_proxy, set_proxy_value

from ..icons import DigitIcons, SelectionIcons, TextIcons

NUM_OPTIONS = [1, 2, 3, 4]
TEXT_OPTIONS = ["foo", "bar", "baz", "qux"]

ICON_ITEM_PATH = "karabogui.controllers.icons_dialogs.IconItem"


class NumberObject(Configurable):
    prop = Int32(options=NUM_OPTIONS)


class StringObject(Configurable):
    prop = String(options=TEXT_OPTIONS)


def _digits_model():
    values = [IconData(value=2, equal=True), IconData(value=4, equal=False)]
    return DigitIconsModel(values=values)


def _selection_model():
    values = [IconData(value=i) for i in TEXT_OPTIONS[:2]]
    return SelectionIconsModel(values=values)


def _text_model():
    values = [IconData(value=i) for i in TEXT_OPTIONS[:2]]
    return TextIconsModel(values=values)


def _assert_current_item(controller, proxy, mocker, value,
                         valid=True, changed=True, digit_icon=False):
    # Save a reference on the previous item
    current_item = controller.current_item

    # Set the proxy with mocked icon item to avoid pixmap problems
    method_path = ICON_ITEM_PATH + "._load_pixmap"
    mocker.patch(method_path, return_value=QPixmap())
    set_proxy_value(proxy, "prop", value)

    if not valid:
        assert controller.current_item is None
        return

    if changed:
        assert current_item is not controller.current_item
    else:
        assert current_item is controller.current_item

    current_item = controller.current_item
    assert current_item is not None

    if digit_icon:
        compare = operator.le if current_item.equal else operator.lt
        assert compare(value, float(current_item.value))
    else:
        assert value == current_item.value


def test_digit_icons(gui_app, mocker):
    # set up
    schema = NumberObject.getClassSchema()
    proxy = get_class_property_proxy(schema, "prop")
    controller = DigitIcons(proxy=proxy, model=_digits_model())
    controller.create(None)
    assert controller.widget is not None
    assert controller.current_item is None

    # set values
    _assert_current_item(controller, proxy, mocker, 2, changed=True,
                         digit_icon=True)
    _assert_current_item(controller, proxy, mocker, 1, changed=False,
                         digit_icon=True)
    _assert_current_item(controller, proxy, mocker, 3, changed=True,
                         digit_icon=True)
    _assert_current_item(controller, proxy, mocker, 4, valid=False,
                         digit_icon=True)

    # teardown
    controller.destroy()
    assert controller.widget is None


def test_selection_icons(gui_app, mocker):
    # setup
    schema = StringObject.getClassSchema()
    proxy = get_class_property_proxy(schema, "prop")
    controller = SelectionIcons(proxy=proxy, model=_selection_model())
    controller.create(None)
    assert controller.widget is not None
    assert controller.current_item is None

    # set values
    _assert_current_item(controller, proxy, mocker, "foo", changed=True)
    _assert_current_item(controller, proxy, mocker, "foo", changed=False)
    _assert_current_item(controller, proxy, mocker, "bar", changed=True)

    # widget will automatically create the items for each option
    _assert_current_item(controller, proxy, mocker, "qux", changed=True)

    # teardown
    controller.destroy()
    assert controller.widget is None


def test_text_icons(gui_app, mocker):
    # setup
    schema = StringObject.getClassSchema()
    proxy = get_class_property_proxy(schema, "prop")
    controller = TextIcons(proxy=proxy, model=_text_model())
    controller.create(None)
    assert controller.widget is not None

    # set value
    _assert_current_item(controller, proxy, mocker, "foo", changed=True)
    _assert_current_item(controller, proxy, mocker, "foo", changed=False)
    _assert_current_item(controller, proxy, mocker, "bar", changed=True)

    # widget will NOT automatically create the items for each option
    _assert_current_item(controller, proxy, mocker, "qux", valid=False)

    # teardown
    controller.destroy()
    assert controller.widget is None



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import html
import re

import natsort

from karabo.native import Hash, KaraboError

MESSAGE_TAG = "Message...........:"
PYTHON_TRACEBACK_HEADER = "Traceback (most recent call last):"
MESSAGE_PATTERN = (r"(?<=\n)\s*(?:{message_tag})\s*(.*?)(?=\n|$)"
                   .format(message_tag=MESSAGE_TAG))
MESSAGE_REGEX = re.compile(MESSAGE_PATTERN)
EXCEPTION_REGEX = re.compile(r"(?<=\n)(.*?(?:Exception|Error)):"
                             r"\s*(.*?)(?=\n|$)")
MAX_CHARS = 150


def get_error_message(failure_reason):
    """Reason is a detailed traceback of the failure. In C++ devices, it could
    have a message string. But in Python devices, it could just have an
    exception message. We try to parse it (by regex)
    """

    # 0. Return the whole message if it's short
    if len(failure_reason) <= MAX_CHARS:
        return failure_reason

    # 1. Check if there's a verbose error message from the C++ devices
    messages = MESSAGE_REGEX.findall(failure_reason)
    verbose_message = None
    # We are more interested on the inner error message than the first ones.
    for message in reversed(messages):
        if not message.startswith(PYTHON_TRACEBACK_HEADER):
            verbose_message = message.lstrip()
            break

    # We didn't succeed on getting a verbose message. Now we parse from the
    # exception messages.
    if verbose_message is None:
        python_exceptions = EXCEPTION_REGEX.findall(failure_reason)
        # We are more interested on the inner error message as this provides
        # the significant message.
        for exception, message in reversed(python_exceptions):
            if KaraboError.__name__ not in exception:
                verbose_message = f"{exception}: {message}"
                break
        else:
            if python_exceptions:
                # Only KaraboErrors, show it!
                verbose_message = f"{message}"
            else:
                # XXX: No exception found ...
                verbose_message = "Unknown exception"

    return html.escape(verbose_message)


def realign_topo_hash(topo_hash, attr):
    """Realign the topo hash `topo_hash` according to the attribute `attr`"""
    sorted_attrs = natsort.natsorted(
        [(k, v, a) for k, v, a in topo_hash.iterall()],
        key=lambda ele: ele[2].get(attr))
    ret = Hash()
    for k, v, a in sorted_attrs:
        ret.setElement(k, v, a)

    return ret


def _vigenere_shift(char: str, key_char: str, decrypt: bool = False) -> str:
    if char.isalpha():
        offset = ord("a") if char.islower() else ord("A")
        key_offset = ord("a") if key_char.islower() else ord("A")
        shift = ord(key_char) - key_offset
        shift = -shift if decrypt else shift
        return chr((ord(char) - offset + shift) % 26 + offset)
    return char


def encrypt(text: str, keyword: str) -> str | None:
    kw = (keyword * (len(text) // len(keyword) + 1))[:len(text)]
    return "".join(_vigenere_shift(char, key_char) for char, key_char in
                   zip(text, kw))


def decrypt(text: str, keyword: str) -> str | None:
    kw = (keyword * (len(text) // len(keyword) + 1))[:len(text)]
    return "".join(
        _vigenere_shift(char, key_char, decrypt=True) for char, key_char in
        zip(text, kw))



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on October 26, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy.QtGui import QIcon
from qtpy.QtWidgets import QAction
from traits.api import Bool, Callable, HasStrictTraits, Instance, String


class KaraboAction(HasStrictTraits):
    """Base class for actions"""
    # The icon for this action
    icon = Instance(QIcon)
    # The text label for the action
    text = String
    # The tooltip text shown when hovering over the action
    tooltip = String
    # Whether or not this action is checkable
    checkable = Bool(False)
    # Whether or not this action is checked
    is_checked = Bool(False)
    # Defines the method which is called whenever the action is triggered
    triggered = Callable
    # Object name which can be used for finding actions programmatically
    name = String


def build_qaction(karabo_action, parent):
    """ Create a QAction from given ``karabo_action``

    NOTE: since the callback for the ``triggered`` Callable should be defined
    where it is needed and might have different signatures, the actual
    ``connect`` of a QAction should be done after calling this function.

    :param karabo_action: An object of type ``KaraboAction``
    :param parent: The parent of the ``QAction`` of type ``QObject``
    :return: ``QAction`` objects based on given ``karabo_action``
    """
    q_action = QAction(karabo_action.text, parent)
    # Attach ``KaraboAction`` object to its respective QAction
    q_action.setData(karabo_action)
    if karabo_action.icon is not None:
        q_action.setIcon(karabo_action.icon)
    q_action.setCheckable(karabo_action.checkable)
    if karabo_action.checkable:
        q_action.setChecked(karabo_action.is_checked)
    q_action.setStatusTip(karabo_action.tooltip)
    q_action.setToolTip(karabo_action.tooltip)
    if karabo_action.name:
        # Only if name is not empty
        q_action.setObjectName(karabo_action.name)
    return q_action



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np
from pyqtgraph import ROI, Point, TextItem, functions as fn
from qtpy.QtCore import QPoint, Qt, Slot
from qtpy.QtGui import QTransform
from qtpy.QtWidgets import QAction, QInputDialog, QLabel, QMenu, QWidgetAction

from karabogui import icons
from karabogui.graph.common.utils import float_to_string

from .utils import set_roi_html


class KaraboROI(ROI):

    def __init__(self, pos=(0, 0), size=Point(1, 1), name='',
                 scaleSnap=False, translateSnap=False, pen=None,
                 parent=None):
        super().__init__(pos, size,
                         scaleSnap=scaleSnap,
                         translateSnap=translateSnap,
                         pen=pen, removable=True, parent=parent)
        self.name = name
        self.setZValue(100)
        self._selected = False
        self._scaling = np.array([1, 1])
        self._translation = np.array([0, 0])

        self._add_handles()

        self.textItem = TextItem(
            html=set_roi_html(name=self.name, center=(0, 0)),
            fill=(0, 0, 0, 50))
        self.textItem.setZValue(99)
        self._change_roi_text_item_details()

        self.sigRegionChanged.connect(self._change_roi_text_item_details)
        self.setAcceptedMouseButtons(Qt.RightButton | Qt.LeftButton)
        self.set_visible(False)

    def _add_handles(self):
        pass

    def mouseClickEvent(self, event):
        super().mouseClickEvent(event)

    # ---------------------------------------------------------------------
    # Public methods

    def update_geometry_from_transform(self, scaling, translation,
                                       update=True):
        self.setPos(self._absolute_position * scaling + translation,
                    update=False)
        self.setSize(self._absolute_size * scaling, update=update)
        self.stateChanged(finish=True)

        self._scaling = scaling
        self._translation = translation
        self._change_roi_text_item_details()

    def set_visible(self, visible=True):
        self.setVisible(visible)
        if self.textItem:
            self.textItem.setVisible(self.selected and visible)

    def redraw(self):
        self.update()

    def _set_handle_visible(self, visible):
        for handle in self.handles:
            handle["item"].setVisible(visible)

    # ---------------------------------------------------------------------
    # PyQt slots

    @Slot()
    def _change_roi_text_item_details(self):
        if not self.textItem:
            return

        direction = np.array([0, 1])  # lower left
        self.textItem.setPos(*(self.pos() + (self.size() * direction)))
        center = [float_to_string(coord) for coord in self.center]
        size = [float_to_string(s) for s in self.size()]
        self.textItem.setHtml(set_roi_html(name=self.name,
                                           center=center,
                                           size=size))

    # ---------------------------------------------------------------------
    # Properties

    @property
    def selected(self):
        return self._selected

    @selected.setter
    def selected(self, selected):
        self._selected = selected

        # Set the visual changes to reflect the change
        width = 2 if selected else 1
        self.pen.setWidthF(width)
        self.currentPen.setWidthF(width)
        self._set_handle_visible(selected)

        # Redraw ROI item
        self.redraw()

        # Then toggle the textItem to show only when ROI selected
        if self.textItem:
            self.textItem.setVisible(selected)

    @property
    def _absolute_position(self):
        return np.array((self.pos() - self._translation) / self._scaling)

    @property
    def _absolute_size(self):
        return np.array(self.size()) / self._scaling

    @property
    def center(self):
        return self.pos() + (self.size() / 2)

    @property
    def coords(self):
        return tuple(self._absolute_position) + tuple(self._absolute_size)

    # ---------------------------------------------------------------------
    # PyQtGraph methods

    def _makePen(self):
        # Generate the pen color for this ROI based on its current state.
        if not self.mouseHovering:
            return self.pen

        pen = fn.mkPen(255, 255, 0)
        pen.setWidthF(3)
        return pen

    def raiseContextMenu(self, ev):
        menu = self.getMenu()
        pos = ev.screenPos()
        menu.popup(QPoint(int(pos.x()), int(pos.y())))

    def getMenu(self):
        if self.menu is not None:
            self.menu.destroy()

        parent = self.parent()
        self.menu = menu = QMenu(parent)
        menu.setTitle("ROI")
        # Label action
        label = QLabel(self.name or "Region of Interest")
        label.setMargin(2)
        label.setAlignment(Qt.AlignCenter)
        label_action = QWidgetAction(menu)
        label_action.setDefaultWidget(label)
        menu.addAction(label_action)
        menu.addSeparator()
        # Remove action
        remove_action = QAction("Remove", menu)
        remove_action.triggered.connect(self.removeClicked)
        remove_action.setIcon(icons.delete)
        menu.addAction(remove_action)
        # Configure action
        configure_action = QAction("Configure", menu)
        configure_action.triggered.connect(self._configure_roi)
        configure_action.setIcon(icons.edit)
        menu.addAction(configure_action)

        return menu

    def mouseDragEvent(self, ev):
        if ev.isStart():
            if ev.button() == Qt.LeftButton:
                # NOTE: We disable the autorange on ROI movement to make sure
                # it does not get out of bounds!
                self.getViewBox().disableAutoRange()
                self.setSelected(True)
                self.setCursor(Qt.ClosedHandCursor)
                self.sigRegionChangeStarted.emit(self)
                if self.translatable:
                    self.isMoving = True
                    self.preMoveState = self.getState()
                    self.cursorOffset = self.pos() - self.mapToParent(
                        ev.buttonDownPos())
                    ev.accept()
                else:
                    ev.ignore()

        elif ev.isFinish():
            if self.translatable:
                if self.isMoving:
                    self.stateChangeFinished()
                self.isMoving = False
            self.setCursor(Qt.OpenHandCursor)
            return

        if (self.translatable and self.isMoving
                and ev.buttons() == Qt.LeftButton):
            snap = True if (ev.modifiers() & Qt.ControlModifier) else None
            newPos = self.mapToParent(ev.pos()) + self.cursorOffset
            self.translate(newPos - self.pos(), snap=snap, finish=False)

    def movePoint(self, handle, pos, modifiers=None, finish=True,
                  coords='parent'):
        # pos is the new position of the handle in scene coords,
        # as requested by the handle.
        if modifiers is None:
            modifiers = Qt.NoModifier
        newState = self.stateCopy()
        index = self.indexOfHandle(handle)
        h = self.handles[index]
        p0 = self.mapToParent(h['pos'] * self.state['size'])
        p1 = Point(pos)

        if coords == 'parent':
            pass
        elif coords == 'scene':
            p1 = self.mapSceneToParent(p1)
        else:
            raise Exception(
                "New point location must be given in either 'parent' or"
                " 'scene' coordinates.")

        if 'center' in h:
            c = h['center']
            cs = c * self.state['size']
            lp0 = self.mapFromParent(p0) - cs
            lp1 = self.mapFromParent(p1) - cs

        if h['type'] == 't':
            snap = True if (modifiers & Qt.ControlModifier) else None
            self.translate(p1 - p0, snap=snap, update=False)

        elif h['type'] == 'f':
            newPos = self.mapFromParent(p1)
            h['item'].setPos(newPos)
            h['pos'] = newPos
            self.freeHandleMoved = True

        elif h['type'] == 's':
            # If a handle and its center have the same x or y value,
            # we can't scale across that axis.
            if h['center'][0] == h['pos'][0]:
                lp1[0] = 0
            if h['center'][1] == h['pos'][1]:
                lp1[1] = 0

            # ---- Patch ----
            # snap
            if self.scaleSnap or (modifiers & Qt.ControlModifier):
                lp1[0] = np.floor(lp1[0] / self._scaling[0]) * self._scaling[0]
                lp1[1] = np.floor(lp1[1] / self._scaling[1]) * self._scaling[1]

            # ---- Patch End ----

            # preserve aspect ratio (this can override snapping)
            if h['lockAspect'] or (modifiers & Qt.AltModifier):
                # arv = Point(self.preMoveState['size']) -
                lp1 = lp1.proj(lp0)

            # determine scale factors and new size of ROI
            hs = h['pos'] - c
            if hs[0] == 0:
                hs[0] = 1
            if hs[1] == 0:
                hs[1] = 1
            newSize = lp1 / hs

            # Perform some corrections and limit checks
            if newSize[0] == 0:
                newSize[0] = newState['size'][0]
            if newSize[1] == 0:
                newSize[1] = newState['size'][1]
            if not self.invertible:
                if newSize[0] < 0:
                    newSize[0] = newState['size'][0]
                if newSize[1] < 0:
                    newSize[1] = newState['size'][1]
            if self.aspectLocked:
                newSize[0] = newSize[1]

            # Move ROI so the center point occupies the same scene location
            # after the scale
            s0 = c * self.state['size']
            s1 = c * newSize
            cc = self.mapToParent(s0 - s1) - self.mapToParent(Point(0, 0))

            # update state, do more boundary checks
            newState['size'] = newSize
            newState['pos'] = newState['pos'] + cc
            if self.maxBounds is not None:
                r = self.stateRect(newState)
                if not self.maxBounds.contains(r):
                    return

            self.setPos(newState['pos'], update=False)
            self.setSize(newState['size'], update=False)

        elif h['type'] in ['r', 'rf']:
            if h['type'] == 'rf':
                self.freeHandleMoved = True

            if not self.rotatable:
                return
            # If the handle is directly over its center point, we can't
            # compute an angle.
            try:
                if lp1.length() == 0 or lp0.length() == 0:
                    return
            except OverflowError:
                return

            # determine new rotation angle, constrained if necessary
            ang = newState['angle'] - lp0.angle(lp1)
            if ang is None:  # this should never happen..
                return
            if self.rotateSnap or (modifiers & Qt.ControlModifier):
                ang = round(ang / self.rotateSnapAngle) * self.rotateSnapAngle

            # create rotation transform
            tr = QTransform()
            tr.rotate(ang)

            # move ROI so that center point remains stationary after rotate
            cc = self.mapToParent(cs) - (tr.map(cs) + self.state['pos'])
            newState['angle'] = ang
            newState['pos'] = newState['pos'] + cc

            # check boundaries, update
            if self.maxBounds is not None:
                r = self.stateRect(newState)
                if not self.maxBounds.contains(r):
                    return
            self.setPos(newState['pos'], update=False)
            self.setAngle(ang, update=False)

            # If this is a free-rotate handle, its distance from the center
            # may change.

            if h['type'] == 'rf':
                h['item'].setPos(self.mapFromScene(
                    p1))  # changes ROI coordinates of handle
                h['pos'] = self.mapFromParent(p1)

        elif h['type'] == 'sr':
            try:
                if lp1.length() == 0 or lp0.length() == 0:
                    return
            except OverflowError:
                return

            ang = newState['angle'] - lp0.angle(lp1)
            if ang is None:
                return
            if self.rotateSnap or (modifiers & Qt.ControlModifier):
                ang = round(ang / self.rotateSnapAngle) * self.rotateSnapAngle

            # --- Patch ---

            if self.aspectLocked or h['center'][0] != h['pos'][0]:
                newState['size'][0] = self.state['size'][
                                          0] * lp1.length() / lp0.length()
                if self.scaleSnap:  # use CTRL only for angular snap here.
                    newState['size'][0] = (
                            np.floor(newState['size'][0]
                                     / self._scaling[0])
                            * self._scaling[0])

            if self.aspectLocked or h['center'][1] != h['pos'][1]:
                newState['size'][1] = self.state['size'][
                                          1] * lp1.length() / lp0.length()
                if self.scaleSnap:  # use CTRL only for angular snap here.
                    newState['size'][1] = (
                            np.floor(newState['size'][1]
                                     / self._scaling[1])
                            * self._scaling[0])

            # --- Patch End ---

            if newState['size'][0] == 0:
                newState['size'][0] = 1
            if newState['size'][1] == 0:
                newState['size'][1] = 1

            c1 = c * newState['size']
            tr = QTransform()
            tr.rotate(ang)

            cc = self.mapToParent(cs) - (tr.map(c1) + self.state['pos'])
            newState['angle'] = ang
            newState['pos'] = newState['pos'] + cc
            if self.maxBounds is not None:
                r = self.stateRect(newState)
                if not self.maxBounds.contains(r):
                    return

            self.setState(newState, update=False)

        self.stateChanged(finish=finish)

    # ---------------------------------------------------------------------
    # Private methods

    @Slot()
    def _configure_roi(self):
        text, ok = QInputDialog.getText(None, 'Configure ROI', 'Name:',
                                        text=self.name)
        if ok:
            self.name = text
            self._change_roi_text_item_details()



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# Created on January 24, 2019
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################

from qtpy.QtCore import Qt
from qtpy.QtGui import QFont
from qtpy.QtWidgets import QFrame, QLabel
from traits.api import Instance

from karabo.common.scenemodel.api import WidgetNodeModel
from karabogui.binding.api import WidgetNodeBinding
from karabogui.const import WIDGET_MIN_HEIGHT
from karabogui.controllers.api import (
    BaseBindingController, register_binding_controller)
from karabogui.indicators import ALL_OK_COLOR
from karabogui.util import generateObjectName


@register_binding_controller(ui_name='Widget Node Field',
                             klassname='WidgetNode',
                             binding_type=WidgetNodeBinding,
                             priority=90)
class DisplayWidgetNode(BaseBindingController):
    # The scene data model class for this controller
    model = Instance(WidgetNodeModel, args=())

    def create_widget(self, parent):
        widget = QLabel(parent)
        widget.setMinimumHeight(WIDGET_MIN_HEIGHT)
        widget.setWordWrap(True)
        widget.setAlignment(Qt.AlignCenter)

        objectName = generateObjectName(self)
        widget.setObjectName(objectName)
        style_sheet = (f"QWidget#{objectName}" +
                       f" {{ background-color : rgba{ALL_OK_COLOR}; }}")
        widget.setStyleSheet(style_sheet)
        widget.setFrameStyle(QFrame.Box)
        widget.setFont(QFont("Times", 8, QFont.Cursive))

        return widget

    def binding_update(self, proxy):
        info = proxy.binding.displayType.split('|')
        # The real node type is mangled in the displayType
        node_type = info[1] if len(info) > 1 else "NaN"
        template = f"WidgetNode\nNodeType: {node_type}"
        self.widget.setText(template)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from unittest import mock

from karabogui.testing import GuiTestCase
from karabogui.widgets.find_toolbar import FindToolBar


class TestFindToolBar(GuiTestCase):

    def setUp(self):
        super().setUp()
        self.find_toolbar = FindToolBar()
        self.find_toolbar.find_line_edit.setText("foo")

    def test_find(self):
        """Find should emit the 'findRequest' signal with correct arguments"""
        mock_slot = mock.Mock()
        self.find_toolbar.findRequested.connect(mock_slot)

        # Find Next
        self.find_toolbar.findNext()
        assert mock_slot.call_count == 1
        call_args = mock_slot.call_args[0]
        assert call_args == ("foo", False, False)

        # Find Previous
        mock_slot.reset_mock()
        self.find_toolbar.findPrevious()
        assert mock_slot.call_count == 1
        call_args = mock_slot.call_args[0]
        assert call_args == ("foo", False, True)

        # Find Next case-sensitive
        self.find_toolbar.match_case.setChecked(True)
        mock_slot.reset_mock()
        self.find_toolbar.findNext()
        assert mock_slot.call_count == 1
        call_args = mock_slot.call_args[0]
        assert call_args == ("foo", True, False)

    def test_setResultText(self):
        """Test the label gets updated with the correct text"""
        self.find_toolbar.setResultText(0)
        assert self.find_toolbar.result_label.text() == "0 Results"
        self.find_toolbar.setResultText(10)
        assert self.find_toolbar.result_label.text() == "10 Results"

        self.find_toolbar.setResultText(1)
        assert self.find_toolbar.result_label.text() == "1 Result"



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy import uic
from qtpy.QtWidgets import QDialog

from .utils import get_dialog_ui


class RequestTimeDialog(QDialog):
    """The RequestTime Dialog class that provides a time range"""
    def __init__(self, start, end, parent=None):
        super().__init__(parent)
        self.setModal(False)
        uic.loadUi(get_dialog_ui("request_time_dialog.ui"), self)
        self.dt_start.setDateTime(start)
        self.dt_end.setDateTime(end)

    def get_start_and_end_time(self):
        """Return the start and end time in seconds"""
        start = self.dt_start.dateTime().toMSecsSinceEpoch() / 1000
        end = self.dt_end.dateTime().toMSecsSinceEpoch() / 1000
        return start, end



#############################################################################
# Author: <john.wiggins@xfel.eu>
# Created on November 23, 2017
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy.QtCore import QSize, Qt, Slot
from qtpy.QtGui import QFontMetrics
from qtpy.QtWidgets import QAction, QDialog, QFrame, QLabel, QSizePolicy

from karabogui.dialogs.textdialog import TextDialog
from karabogui.fonts import get_qfont
from karabogui.widgets.hints import KaraboSceneWidget

MARGINS = (5, 0, 5, 0)  # left, top, right, bottom
WIDTH_MARGIN = MARGINS[0] + MARGINS[2]
HEIGHT_MARGIN = MARGINS[1] + MARGINS[3]


class LabelWidget(KaraboSceneWidget, QLabel):
    """A label which can appear in a scene
    """

    def __init__(self, model, parent=None):
        super().__init__(model.text, model=model, parent=parent)
        self.setSizePolicy(QSizePolicy.MinimumExpanding,
                           QSizePolicy.Minimum)
        self.setFrameShape(QFrame.Box)
        self.fm = None
        self.set_model(model)

        # Base edit action
        edit_action = QAction("Edit Label", self)
        edit_action.triggered.connect(self.edit_colors_text)
        self.addAction(edit_action)

    def set_model(self, model):
        """Set the new ``model`` and update the widget properties.
        """
        self.model.trait_set(text=model.text, frame_width=model.frame_width,
                             font=model.font, background=model.background,
                             foreground=model.foreground,
                             alignh=model.alignh)
        font = get_qfont(model.font)
        self.setFont(font)
        self.setAlignment(Qt.AlignmentFlag(model.alignh) | Qt.AlignVCenter)
        self.fm = QFontMetrics(font)
        self.setLineWidth(model.frame_width)
        self.setToolTip(model.text)

        # Set the stylesheet for background and foreground color
        sheet = []
        sheet.append(f"color: {model.foreground};")
        if model.background:
            sheet.append(f"background-color: {model.background};")
        self.setStyleSheet(f"QLabel {{ {''.join(sheet)} }}")
        self.setGeometry(model.x, model.y, model.width, model.height)
        self.create_display_text()

    def minimumSizeHint(self):
        """Reimplemented `sizeHint` of KaraboSceneWidget"""
        fm = QFontMetrics(self.font())
        width = fm.width(self.model.text)
        height = fm.height() + self.model.frame_width
        size = QSize(width + WIDTH_MARGIN, max(height, 20) + HEIGHT_MARGIN)

        return size

    def sizeHint(self):
        """Reimplemented `sizeHint` of KaraboSceneWidget since we want to get
        the suggested sizeHint from the font metrics and not from `QLabel`

        Calculate the size hint from the text if model is newly
        instantiated (no width/height yet). Else, return the model size."""
        size = QSize()
        model = self.model
        if (model.x, model.y) != (0, 0):
            size = QSize(model.width, model.height)
        if size.isEmpty():
            fm = QFontMetrics(self.font())
            width = fm.width(self.model.text)
            height = fm.height() + self.model.frame_width
            size = QSize(width + WIDTH_MARGIN, max(height, 20) + HEIGHT_MARGIN)
        return size

    def resizeEvent(self, event):
        """Reimplemented function of PyQt for adjust the display text"""
        super().resizeEvent(event)
        self.create_display_text()

    @property
    def text_width(self):
        return self.width() - WIDTH_MARGIN

    def create_display_text(self):
        """Calculate the effective display text

        Use the font metric to calculate the width of the text. Is the width
        is larger than the widget size, the text is returned as elided.
        For synchronization, this method is invoked on resizeEvents as well.
        """
        display = self.text()
        text = self.model.text
        text_width = self.text_width
        if self.fm.width(text) > text_width:
            text = self.fm.elidedText(text, Qt.ElideRight, text_width)
        if display != text:
            self.setText(text)

    def add_proxies(self, proxies):
        """Satisfy the informal widget interface."""

    def apply_changes(self):
        """Satisfy the informal widget interface."""

    def decline_changes(self):
        """Satisfy the informal widget interface."""

    def destroy(self):
        """Satisfy the informal widget interface."""
        super().destroy()

    def set_visible(self, visible):
        """Satisfy the informal widget interface."""

    def update_global_access_level(self, level):
        """Satisfy the informal widget interface."""

    def set_geometry(self, rect):
        self.model.trait_set(x=rect.x(), y=rect.y(),
                             width=rect.width(), height=rect.height())
        self.setGeometry(rect)

    def translate(self, offset):
        new_pos = self.pos() + offset
        self.model.trait_set(x=new_pos.x(), y=new_pos.y())
        self.move(new_pos)

    @Slot()
    def edit_colors_text(self):
        self.edit(scene_view=None)
        self.update()

    def edit(self, scene_view):
        dialog = TextDialog(self.model, alignment=True, parent=scene_view)
        if dialog.exec() == QDialog.Rejected:
            return

        self.set_model(dialog.label_model)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
# flake8: noqa
from .buttons import PopupButtonWidget
from .container import ControllerContainer
from .grid import GridView
from .label import LabelWidget
from .link import DeviceSceneLinkWidget, SceneLinkWidget, WebLinkWidget
from .renderer import ImageRendererWidget
from .status import InstanceStatusWidget
from .sticker import StickerWidget
from .unknown import UnknownSvgWidget, UnknownWidget



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import contextlib
import os
import sys
import traceback
import unittest
from platform import system
from xml.etree.ElementTree import parse

import pytest
from qtpy.QtCore import Qt
from qtpy.QtTest import QTest
from qtpy.QtWidgets import QApplication

import karabogui.singletons.api as singletons_mod
from karabo.common.api import Capabilities, State
from karabo.common.scenemodel.api import set_scene_reader
from karabo.native import (
    AccessLevel, AccessMode, Configurable, Double, Hash, Int32, String, Unit,
    VectorString)
from karabogui.background import create_background_timer
from karabogui.binding.api import (
    BindingRoot, DeviceClassProxy, DeviceProxy, PropertyProxy, ProxyStatus,
    apply_configuration, build_binding)
from karabogui.controllers.api import populate_controller_registry
from karabogui.util import process_qt_events


class GuiTestCase(unittest.TestCase):
    """ A convenient base class for gui test cases"""

    def setUp(self):
        os.environ["KARABO_TEST_GUI"] = "1"
        if system() == "Darwin" and "QT_MAC_WANTS_LAYER" not in os.environ:
            os.environ["QT_MAC_WANTS_LAYER"] = "1"
        app = QApplication.instance()
        if app is None:
            app = QApplication(sys.argv)
        self.app = app
        # Check organization for QSettings for tests, see the setting in
        # `programs/tests/test_base.py`
        app.setOrganizationName("TestFacility")
        app.setOrganizationDomain("karabo.test.eu")
        app.setApplicationName("pytest")

        create_background_timer()
        # AFTER the QApplication is created!
        import karabogui.access as krb_access
        krb_access.GLOBAL_ACCESS_LEVEL = AccessLevel.OPERATOR
        from karabogui import icons
        icons.init()
        populate_controller_registry()

    def tearDown(self):
        pass

    def process_qt_events(self, ms=10):
        # Give the event loop 10ms to process its events
        process_qt_events(timeout=ms)

    def click(self, button_widget, button=Qt.LeftButton):
        QTest.mouseClick(button_widget, button)


def click_button(button_widget, button=Qt.LeftButton):
    QTest.mouseClick(button_widget, button)


def keySequence(widget, sequence):
    QTest.keySequence(widget, sequence)


@contextlib.contextmanager
def assert_trait_change(obj, name):
    """A context manager which watches for traits events
    """
    events = []

    def handler():
        events.append(None)

    obj.on_trait_change(handler, name)
    try:
        yield
    finally:
        obj.on_trait_change(handler, name, remove=True)
        if len(events) == 0:
            msg = "Expected change for trait '{}' was not observed!"
            raise AssertionError(msg.format(name))


@contextlib.contextmanager
def flushed_registry():
    """Avoid polluting the global binding controller registry with
    test controller classes
    """
    from karabogui.controllers import registry as registrymod

    controller_registry = registrymod._controller_registry.copy()
    registrymod._controller_registry.clear()
    model_registry = registrymod._controller_models.copy()
    registrymod._controller_models.clear()

    yield

    registrymod._controller_registry = controller_registry
    registrymod._controller_models = model_registry


def get_property_proxy(schema, name, device_id="TestDevice"):
    """Given a device schema and a property name, return a complete
    PropertyProxy object with a `root_proxy` of type `DeviceProxy`.
    """
    binding = BindingRoot() if schema is None else build_binding(schema)
    status = (ProxyStatus.ONLINEREQUESTED if schema is None
              else ProxyStatus.SCHEMA)
    root_proxy = DeviceProxy(binding=binding, device_id=device_id)
    root_proxy.status = status
    return PropertyProxy(root_proxy=root_proxy, path=name)


def get_class_property_proxy(schema, name):
    """Given a device schema and a property name, return a complete
    PropertyProxy object with a `root_proxy` of type `DeviceClassProxy`.
    """
    binding = build_binding(schema)
    root = DeviceClassProxy(binding=binding, server_id="Test",
                            status=ProxyStatus.OFFLINE)
    return PropertyProxy(root_proxy=root, path=name)


def set_proxy_value(proxy, name, value):
    """Use `apply_configuration` to set a single property value on a proxy
    """
    apply_configuration(Hash(name, value), proxy.root_proxy.binding)


def set_proxy_hash(proxy, hsh, timestamp=None):
    """Use `apply_configuration` to set a a hash on a proxy

    :param timestamp: Optionally a timestamp can be provided for the hash
    """
    if timestamp is not None:
        timestamp.toHashAttributes(hsh)

    apply_configuration(hsh, proxy.root_proxy.binding)


@contextlib.contextmanager
def singletons(**objects):
    """Provide a collection of singletons to be used for the duration of a
    `with`-block.
    """
    singletons_dict = singletons_mod.__singletons
    # Remember what got replaced
    replaced = {k: singletons_dict[k] for k in objects if k in singletons_dict}
    try:
        # Replace and yield
        for key, obj in objects.items():
            singletons_dict[key] = obj
        yield
    finally:
        # Put things back as they were
        for key, obj in replaced.items():
            singletons_dict[key] = obj
        for key in objects:
            if key not in replaced:
                del singletons_dict[key]


@contextlib.contextmanager
def assert_no_throw():
    """A simple context manager to assert a no-throw"""
    try:
        yield None
    except Exception as exception:
        trace = ''.join(traceback.format_exception(
            type(exception), exception, exception.__traceback__))
        raise pytest.fail(f"Did raise unexpected exception: {trace}")


@contextlib.contextmanager
def access_level(level):
    """Temporary move the access level to `level`"""
    import karabogui.access as krb_access
    old_level = krb_access.GLOBAL_ACCESS_LEVEL
    try:
        krb_access.GLOBAL_ACCESS_LEVEL = level
        yield
    finally:
        krb_access.GLOBAL_ACCESS_LEVEL = old_level


DEVICE_CAPA = Capabilities.PROVIDES_SCENES + Capabilities.PROVIDES_INTERFACES


def system_hash():
    """Generate a system hash which will be built into a system tree
    """
    h = Hash()

    h["server.swerver"] = None
    h["server.swerver", ...] = {
        "host": "BIG_IRON",
        "visibility": AccessLevel.OBSERVER,
        "deviceClasses": ["FooClass", "BarClass"],
        "type": "server",
        "visibilities": [AccessLevel.OBSERVER, AccessLevel.OBSERVER]
    }
    h["device.divvy"] = None
    h["device.divvy", ...] = {
        "host": "BIG_IRON",
        "archive": True,
        "visibility": AccessLevel.OBSERVER,
        "type": "device",
        "capabilities": DEVICE_CAPA,
        "serverId": "swerver",
        "classId": "FooClass",
        "status": "ok",
        "interfaces": 0,
    }
    h["macro.macdonald"] = None
    h["macro.macdonald", ...] = {
        "host": "BIG_IRON",
        "archive": False,
        "visibility": AccessLevel.OBSERVER,
        "serverId": "swerver",
        "classId": "BarClass",
        "capabilities": 0,
        "status": "ok",
        "interfaces": 0,
    }

    h["device.orphan"] = None
    h["device.orphan", ...] = {
        "host": "BIG_IRON",
        "visibility": AccessLevel.OBSERVER,
        "archive": True,
        "serverId": "__none__",
        "classId": "Parentless",
        "capabilities": 0,
        "status": "ok",
        "interfaces": 0,
    }

    h["client.charlie"] = None
    h["client.charlie", ...] = {
        "host": "BIG_IRON",
        "visibility": AccessLevel.OBSERVER,
        "archive": False,
        "serverId": "__none__",
        "classId": "NoClass",
        "status": "ok",
    }

    return h


def development_system_hash():
    h = Hash()

    h["server.devserver"] = None
    h["server.devserver", ...] = {
        "host": "BIG_IRON",
        "visibility": AccessLevel.OBSERVER,
        "deviceClasses": ["MetaMacro"],
        "type": "server",
        "lang": "macro",
        "serverFlags": 1,
        "visibilities": [AccessLevel.OBSERVER]
    }

    h["server.stableserver"] = None
    h["server.stableserver", ...] = {
        "host": "BIG_IRON",
        "visibility": AccessLevel.OBSERVER,
        "deviceClasses": ["MetaMacro"],
        "type": "server",
        "lang": "macro",
        "serverFlags": 0,
        "visibilities": [AccessLevel.OBSERVER]
    }

    return h


def device_hash():
    """Generate a device hash which will be built into a device tree
    """
    h = Hash()

    h["device.XFEL/FOO/1"] = None
    h["device.XFEL/FOO/1", ...] = {
        "host": "BIG_IRON",
        "archive": False,
        "visibility": AccessLevel.OBSERVER,
        "capabilities": DEVICE_CAPA,
        "serverId": "swerver",
        "classId": "FooClass",
        "status": "ok",
        "interfaces": 0,
    }

    h["device.XFEL/FOO/2"] = None
    h["device.XFEL/FOO/2", ...] = {
        "host": "BIG_IRON",
        "archive": True,
        "visibility": AccessLevel.OBSERVER,
        "capabilities": DEVICE_CAPA,
        "serverId": "swerver",
        "classId": "FooClass",
        "status": "ok",
        "interfaces": 0,
    }

    h["device.XFEL/BAR/1"] = None
    h["device.XFEL/BAR/1", ...] = {
        "host": "BIG_IRON",
        "archive": True,
        "visibility": AccessLevel.OBSERVER,
        "capabilities": DEVICE_CAPA,
        "serverId": "swerver",
        "classId": "BarClass",
        "status": "error",
        "interfaces": 0,
    }

    return h


def system_hash_server_and_plugins():
    """Generate a system hash which will be built into a system tree
    """
    h = Hash()

    h["server.myserver"] = None
    h["server.myserver", ...] = {
        "host": "exflpxc_something",
        "visibility": AccessLevel.OBSERVER,
        "deviceClasses": ["FooClass", "BarClass"],
        "visibilities": [AccessLevel.OBSERVER, AccessLevel.OBSERVER]
    }
    h["server.samedeviceclasses"] = None
    h["server.samedeviceclasses", ...] = {
        "host": "exflpxc_something",
        "visibility": AccessLevel.EXPERT,
        "deviceClasses": ["FooClass", "BlahClass", "HooClass_0"],
        "visibilities": [AccessLevel.OBSERVER, AccessLevel.OBSERVER,
                         AccessLevel.OBSERVER]
    }

    h["server.differentaccesslevel"] = None
    h["server.differentaccesslevel", ...] = {
        "host": "exflpxc_something",
        "visibility": AccessLevel.EXPERT,
        "deviceClasses": ["FooClass", "BarClass", "HooClass_1"],
        "visibilities": [AccessLevel.OBSERVER, AccessLevel.EXPERT,
                         AccessLevel.OBSERVER]
    }

    h["device.orphan"] = None
    h["device.orphan", ...] = {
        "host": "exflpxc_something",
        "visibility": AccessLevel.OBSERVER,
        "serverId": "myserver",
        "classId": "FooClass",
        "status": "ok",
        "interfaces": 0,
        "capabilities": 0,
    }

    h["device.eddie"] = None
    h["device.eddie", ...] = {
        "host": "exflpxc_something",
        "visibility": AccessLevel.OBSERVER,
        "serverId": "myserver",
        "classId": "BarClass",
        "status": "ok",
        "interfaces": 0,
        "capabilities": 0,
    }

    return h


class SimpleDeviceSchema(Configurable):
    """A simple device schema"""
    state = String(
        defaultValue=State.ON,
        displayType="State",
        enum=State)
    doubleProperty = Double(
        defaultValue=None,
        unitSymbol=Unit.METER)
    stringProperty = String(
        defaultValue="foo",
        options=["foo", "bar", "baz", "qux"])
    readOnlyProperty = Int32(
        defaultValue=0,
        accessMode=AccessMode.READONLY)
    availableScenes = VectorString(
        defaultValue=["scene"])


class DeviceSchemaAllowedState(SimpleDeviceSchema):
    """Device Schema with allowed state defined for an integer property"""
    intProperty = Int32(
        defaultValue=10,
        unitSymbol=Unit.METER,
        allowedStates={State.PASSIVE})

    initOnlyString = String(
        displayedName="Init Only String",
        defaultValue="Karabo",
        accessMode=AccessMode.INITONLY, )


def get_device_schema():
    return SimpleDeviceSchema.getClassSchema()


def get_device_schema_allowed_state():
    return DeviceSchemaAllowedState.getClassSchema()


def check_renderer_against_svg(renderer, svgfile):
    """Yeah... Check an SVG for elements which we suspect it should have.
    """

    def _get_path_ids(filename):
        tree = parse(filename)
        ids = []
        for elem in tree.iter('{http://www.w3.org/2000/svg}path'):
            ids.append(elem.get('id'))
        return ids

    for name in _get_path_ids(svgfile):
        assert renderer.elementExists(name)


@contextlib.contextmanager
def lazy_scene_reading():
    try:
        set_scene_reader(lazy=True)
        yield
    finally:
        set_scene_reader(lazy=False)


def set_test_organization_info(app):
    """Set the organization info for the application, to isolate the
    QSettings for tests"""
    app.setOrganizationName("TestFacility")
    app.setOrganizationDomain("karabo.test.eu")
    app.setApplicationName("pytest")



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from unittest import main, mock

from qtpy.QtCore import Qt
from qtpy.QtGui import QBrush, QColor, QPen

from karabogui.dialogs.api import PenDialog
from karabogui.testing import GuiTestCase


class TestLoginDialog(GuiTestCase):

    def test_basic_dialog(self):
        """Check the basics of the pen dialog"""
        # 1. First initialize with pen and brush, brush is solid
        pen = QPen()
        pen.setColor(QColor(0, 255, 0))
        pen.setWidth(2)
        brush = QBrush()
        brush.setStyle(Qt.SolidPattern)
        dialog = PenDialog(pen, brush)

        assert pen.miterLimit() == dialog.dsbStrokeMiterLimit.value()
        # We have a Qt.SolidPattern
        assert dialog.gbFill.isChecked() is True

        # 2. Check to initialize without brush
        pen = QPen()
        pen.setStyle(Qt.CustomDashLine)
        dialog = PenDialog(pen)
        assert dialog is not None
        assert dialog.pen.style() == Qt.CustomDashLine
        assert dialog.wDashType.penStyle() == 1

        # 3. Initialize with pen and brush, brush is no brush
        pen = QPen()
        pen.setColor(QColor(0, 255, 0))
        pen.setWidth(2)
        brush = QBrush()
        brush.setStyle(Qt.NoBrush)
        dialog = PenDialog(pen, brush)
        assert dialog.gbFill.isChecked() is False

        with mock.patch("karabogui.dialogs.pen_dialogs.QColorDialog") as d:
            brush_color = QColor(0, 0, 10, 255)
            d.getColor.return_value = brush_color
            dialog.on_pbFillColor_clicked()
            assert dialog.brush.color() == brush_color

        with mock.patch("karabogui.dialogs.pen_dialogs.QColorDialog") as d:
            pen_color = QColor(0, 255, 0, 255)
            d.getColor.return_value = pen_color
            dialog.on_pbStrokeColor_clicked()
            assert dialog.pen.color() == pen_color

        # Note: Check the accept dialog
        # Change to Solid Pattern
        dialog.gbFill.setChecked(True)
        dialog.on_dialog_accepted()
        assert dialog.brush.style() == Qt.SolidPattern
        dialog.sbStrokeWidth.setValue(0)
        dialog.on_dialog_accepted()
        assert dialog.pen.style() == Qt.NoPen
        dialog.sbStrokeWidth.setValue(2)
        dialog.on_dialog_accepted()
        assert dialog.pen.style() != Qt.NoPen
        assert dialog.pen.style() == Qt.SolidLine
        assert dialog.pen.dashOffset() == 0

        dialog.dsbStrokeMiterLimit.setValue(1)
        dialog.on_dialog_accepted()
        assert dialog.pen.miterLimit() == 1


if __name__ == "__main__":
    main()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from abc import abstractmethod

from pyqtgraph import PlotItem, ViewBox
from qtpy.QtCore import QPoint
from qtpy.QtGui import QTransform
from qtpy.QtWidgets import QMenu
from traits.api import ABCHasStrictTraits, Constant, Instance, Property, String

from karabogui.graph.common.api import create_axis_items
from karabogui.graph.common.const import (
    AXIS_ITEMS, AXIS_X, AXIS_Y, ROTATION_FACTOR)
from karabogui.graph.common.enums import AxisType

LABEL_STYLE = {'color': '#000000', 'font-size': '12px'}

SHOWN_AXES = {
    # Normal plot orientation, where x-axis is shown in the bottom
    # and y-axis on the left of the plot.
    "top": ["bottom", "left"],

    # Plot orientation is on the left of the layout, thus the x-axis
    # must be shown on the top and the y-axis on the left of the plot.
    "left": ["right", "top"]
}


class AuxPlotViewBox(ViewBox):
    """ There is a need to subclass the viewbox to prevent the wheel scroll
    from destroying the plot range."""

    def __init__(self, parent=None):
        super().__init__(parent=parent, enableMenu=False)
        self.setMouseEnabled(x=False, y=False)
        self.menu = None

    def set_menu(self, menu):
        self.menu = menu

    def menuEnabled(self):
        return self.menu is not None

    def raiseContextMenu(self, event):
        """Reimplemented function of PyQtGraph"""
        if self.menu is None:
            return
        pos = event.screenPos()
        self.menu.popup(QPoint(int(pos.x()), int(pos.y())))


class AuxPlotItem(PlotItem):

    def __init__(self, orientation="top", axisItems=None, parent=None):
        super().__init__(
            parent=parent, axisItems=axisItems, enableMenu=False,
            viewBox=AuxPlotViewBox(parent=None))

        # Initialize plot item
        if orientation == "left":
            # Add fixed width to not overlap with main plot axes
            self.getAxis("right").setFixedWidth(35)
            # The view and must be rotated.
            self.vb.invertY()
            self.vb.invertX()
        elif orientation == "top":
            # Add fixed width to not overlap with main plot axes
            self.getAxis("bottom").setFixedHeight(20)

        # Polish axes rendering
        for axis in AXIS_ITEMS:
            axis_item = self.getAxis(axis)
            axis_item.setZValue(0)
            axis_item.setVisible(True)
            axis_x = (AXIS_X if orientation in ["top", "bottom"] else AXIS_Y)
            if axis in axis_x:
                axis_item.enableAutoSIPrefix(False)

        self.hideButtons()

    # ---------------------------------------------------------------------
    # Reimplemented `PyQtGraph` methods

    def setLabel(self, axis, text=None, **kwargs):
        super().setLabel(axis, text=text, **LABEL_STYLE)

    def setMenuEnabled(self, enableMenu=True, enableViewBoxMenu=False):
        """Reimplemented function to circumvent behavior of pyqtgraph"""
        self._menuEnabled = enableMenu


class BasePlot(ABCHasStrictTraits):
    plotItem = Instance(AuxPlotItem, clean_up=True)
    orientation = String

    x_label = Constant('')
    y_label = Constant('')

    menu = Property(Instance(QMenu))

    def __init__(self, **traits):
        super().__init__(**traits)

        # Create axis items for different plot orientations
        shown_axes = SHOWN_AXES[self.orientation]
        axis_items = create_axis_items(AxisType.AuxPlot, shown_axes)

        # Instantiate plotItem
        plotItem = AuxPlotItem(orientation=self.orientation,
                               axisItems=axis_items)
        self.plotItem = plotItem

        # Set labels only if there's existing value
        if self.x_label:
            plotItem.setLabel(shown_axes[0], self.x_label)
        if self.y_label:
            plotItem.setLabel(shown_axes[1], self.y_label)

    def _add_plot_item(self):
        item = self.plotItem.plot([], [])

        # Transform added plot item if the orientation is unconventional
        # (left, right, bottom). For now, left is only implemented
        if self.orientation == "left":
            angle = 270
            transform = QTransform()
            transform.rotate(angle)
            transform.scale(*ROTATION_FACTOR[angle])
            item.setTransform(transform)

        return item

    def _set_menu(self, menu):
        self.plotItem.vb.set_menu(menu)

    def destroy(self):
        """Destroy the plotItem cleanly"""
        self.plotItem.deleteLater()

    @abstractmethod
    def set_data(self, x_data, y_data):
        """Used by subclass to tune the data and the graphical parameters
           according to its specifications
           (e.g., histograms have gradient fill)"""

    def clear_data(self):
        for data_item in self.plotItem.dataItems:
            data_item.setData([], [], fillLevel=None, stepMode=False)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from ast import literal_eval

from qtpy.QtCore import Qt
from qtpy.QtGui import QPalette, QValidator
from qtpy.QtWidgets import (
    QDialog, QHBoxLayout, QLineEdit, QToolButton, QWidget)
from traits.api import Instance, Int

from karabo.common.api import KARABO_SCHEMA_REGEX
from karabo.common.scenemodel.api import (
    EditableListModel, EditableRegexListModel)
from karabogui import icons
from karabogui.binding.api import (
    VectorBinding, VectorCharBinding, VectorHashBinding, VectorNoneBinding,
    VectorStringBinding, get_editor_value, get_min_max_size)
from karabogui.const import WIDGET_MIN_HEIGHT
from karabogui.controllers.api import (
    BaseBindingController, ListValidator, is_proxy_allowed,
    register_binding_controller)
from karabogui.dialogs.listedit import ListEditDialog
from karabogui.util import SignalBlocker
from karabogui.validators import RegexListValidator
from karabogui.widgets.hints import LineEdit


class BaseEditableListController(BaseBindingController):
    _internal_widget = Instance(QLineEdit)
    last_cursor_position = Int(0)
    layout = Instance(QHBoxLayout)

    # other internal traits
    _validator = Instance(QValidator)
    _normal_palette = Instance(QPalette)
    _error_palette = Instance(QPalette)

    def create_widget(self, parent):
        widget = QWidget(parent)
        widget.setMinimumHeight(WIDGET_MIN_HEIGHT)
        self.layout = QHBoxLayout(widget)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self._internal_widget = LineEdit(parent)
        self.layout.addWidget(self._internal_widget)
        self._normal_palette = self._internal_widget.palette()
        self._error_palette = QPalette(self._normal_palette)
        self._error_palette.setColor(QPalette.Text, Qt.red)
        widget.setFocusProxy(self._internal_widget)

        self._internal_widget.textChanged.connect(self._on_user_edit)
        self._internal_widget.setFocusPolicy(Qt.StrongFocus)

        text = "Edit"
        tbEdit = QToolButton()
        tbEdit.setStatusTip(text)
        tbEdit.setToolTip(text)
        tbEdit.setIcon(icons.edit)
        tbEdit.setMinimumSize(WIDGET_MIN_HEIGHT, WIDGET_MIN_HEIGHT)
        tbEdit.setMaximumSize(25, 25)
        tbEdit.setFocusPolicy(Qt.NoFocus)
        tbEdit.clicked.connect(self._on_edit_clicked)
        self.layout.addWidget(tbEdit)

        return widget

    def binding_update(self, proxy):
        raise NotImplementedError

    def state_update(self, proxy):
        enable = is_proxy_allowed(proxy)
        self.widget.setEnabled(enable)

    def value_update(self, proxy):
        value = get_editor_value(proxy, [])
        self._set_edit_field_text(value)

    def _on_user_edit(self, text):
        if self.proxy.binding is None:
            return

        acceptable_input = self._internal_widget.hasAcceptableInput()
        self.last_cursor_position = self._internal_widget.cursorPosition()
        if acceptable_input:
            cast = (str if isinstance(self.proxy.binding, VectorStringBinding)
                    else literal_eval)
            self.proxy.edit_value = [cast(val)
                                     for val in text.split(',') if val != '']
        else:
            # erase edit value!
            self.proxy.edit_value = None
        # update color after text change!
        palette = (self._normal_palette if acceptable_input
                   else self._error_palette)
        self._internal_widget.setPalette(palette)

    def on_decline(self):
        if self._internal_widget.palette() == self._error_palette:
            self._internal_widget.setPalette(self._normal_palette)

    def _validate_empty(self):
        """This method validates an empty value on the list
        """
        value = []
        state, _, _ = self._validator.validate(value, 0)
        if state == QValidator.Invalid:
            value = None
        return value

    def _set_edit_field_text(self, value):
        with SignalBlocker(self._internal_widget):
            self._internal_widget.setText(','.join(str(v) for v in value))
        self._internal_widget.setCursorPosition(self.last_cursor_position)

    def _on_edit_clicked(self):
        if self.proxy.binding is None:
            return

        list_edit = ListEditDialog(self.proxy.binding, duplicates_ok=True,
                                   parent=self.widget)
        list_edit.set_list(get_editor_value(self.proxy, []))
        if list_edit.exec() == QDialog.Accepted:
            self.proxy.edit_value = list_edit.values
            with SignalBlocker(self._internal_widget):
                display = ','.join(str(v) for v in list_edit.values)
                self._internal_widget.setText(display)
            self._internal_widget.setCursorPosition(self.last_cursor_position)
        # Give back the focus!
        self._internal_widget.setFocus(Qt.PopupFocusReason)


def _is_compatible(binding):
    """Don't allow editing of goofy vectors"""
    prohibited = (VectorHashBinding, VectorNoneBinding, VectorCharBinding)
    return not isinstance(binding, prohibited)


@register_binding_controller(ui_name='Edit List', is_compatible=_is_compatible,
                             klassname='EditableList', can_edit=True,
                             binding_type=VectorBinding, priority=10)
class EditableList(BaseEditableListController):
    """The editable version of the list widget"""
    model = Instance(EditableListModel, args=())

    def binding_update(self, proxy):
        binding = proxy.binding
        self._validator = ListValidator(binding=binding, parent=self.widget)
        self._internal_widget.setValidator(self._validator)


def _has_regex_attribute(binding):
    attrs = binding.attributes
    return attrs.get(KARABO_SCHEMA_REGEX, None) is not None


@register_binding_controller(ui_name='Edit Regex List', can_edit=True,
                             is_compatible=_has_regex_attribute,
                             klassname='EditableRegexList',
                             binding_type=VectorStringBinding,
                             priority=10)
class EditableRegexList(BaseEditableListController):
    model = Instance(EditableRegexListModel, args=())

    def binding_update(self, proxy):
        binding = proxy.binding
        regex = binding.attributes.get(KARABO_SCHEMA_REGEX, "")
        minSize, maxSize = get_min_max_size(binding)
        self._validator = RegexListValidator(
            regex, minSize=minSize, maxSize=maxSize, parent=self.widget)
        self._internal_widget.setValidator(self._validator)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np
import pytest

from karabo.common.scenemodel.api import VectorBarGraphModel
from karabo.native import Configurable, Hash, Int32, NDArray, VectorInt32
from karabogui.testing import get_class_property_proxy, set_proxy_value

from ..vector_bar_graph import DisplayVectorBarGraph


class Object(Configurable):
    value = VectorInt32()


class NDArrayObject(Configurable):
    value = NDArray(
        defaultValue=np.arange(10),
        shape=(10,),
        dtype=Int32,
    )


@pytest.fixture
def vector_bar_graph_setup(gui_app):
    # setup
    schema = Object.getClassSchema()
    proxy = get_class_property_proxy(schema, "value")
    controller = DisplayVectorBarGraph(proxy=proxy,
                                       model=VectorBarGraphModel())
    controller.create(None)
    assert controller.widget is not None
    yield controller, proxy
    # teardown
    controller.destroy()
    assert controller.widget is None


def test_bar_graph_basics(vector_bar_graph_setup):
    controller, proxy = vector_bar_graph_setup
    value = [2, 4, 6]
    set_proxy_value(proxy, "value", value)
    curve = controller._plot
    assert list(curve.opts.get("x")) == [0, 1, 2]
    assert list(curve.opts.get("height")) == value


def test_bar_graph_width(vector_bar_graph_setup, mocker):
    controller, proxy = vector_bar_graph_setup
    action = controller.widget.actions()[9]
    assert action.text() == "Bar Width"

    dsym = "karabogui.controllers.display.vector_bar_graph.QInputDialog"
    QInputDialog = mocker.patch(dsym)
    QInputDialog.getDouble.return_value = 2.7, True
    action.trigger()
    assert controller.model.bar_width == 2.7
    assert controller._plot.opts["width"] == 2.7


def test_array_bar_graph_basics(gui_app):
    # setup
    schema = NDArrayObject.getClassSchema()
    proxy = get_class_property_proxy(schema, "value")
    controller = DisplayVectorBarGraph(proxy=proxy)
    controller.create(None)
    assert controller.widget is not None

    # test body
    value = np.array(
        [2, 3, 2, 3, 2, 3, 2, 3, 2, 3],
        dtype=np.int32)
    array_hash = Hash("type", 12,
                      "data", value.tobytes())
    set_proxy_value(proxy, "value", array_hash)
    curve = controller._plot
    np.testing.assert_array_equal(
        list(curve.opts.get("x")), np.arange(10))
    np.testing.assert_array_equal(
        list(curve.opts.get("height")), value)

    # teardown
    controller.destroy()
    assert controller.widget is None



#############################################################################
# Author: <john.wiggins@xfel.eu>
# Created on November 3, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################

from qtpy.QtCore import QObject

from karabo.common.api import KARABO_PROJECT_MANAGER, set_modified_flag
from karabo.common.project.api import (
    MemCacheWrapper, get_user_cache, read_lazy_object)
from karabo.common.scenemodel.api import SceneModel
from karabo.native import (
    Hash, get_item_type, read_project_model, write_project_model)
from karabogui import messagebox
from karabogui.events import (
    KaraboEvent, broadcast_event, register_for_broadcasts)
from karabogui.request import get_scene_from_server
from karabogui.singletons.api import get_config, get_network

# This matches the batch size used in the project database
MAX_BUFFER_ITEMS = 50


class ProjectDatabaseConnection(QObject):
    """ An object which handles requests/replies from the GUI server which
    pertain to project data.

    NOTE: Concerning the Cache
    The cache is fundamental to the operation of this database connection. Due
    to the asynchronous nature of reading project data, a cache was
    incorporated to "smooth" things out a bit. In the end, this means that data
    is always delivered to deserialization code from the cache. Structuring
    things this way is a bit complex, but it enables recursively reading
    (with `read_lazy_object`) a project where at any point you might read
    data which is not locally available.

    Just to be clear, the way this works is:
        * An empty object is created with a valid UUID
        * An attempt is made to initialize that object with `read_lazy_object`,
          which is passed a "storage" object which can be a `ProjectDBCache`,
          or an instance of this connection class, or a `MemCacheWrapper` which
          forwards misses to some other storage object.
        * Data which is not immediately available in the storage object is
          requested from the network.
        * Sometime later, data is received asynchronously and written to either
          a memory or disk cache.
        * Objects for which data arrived are read again using
          `read_lazy_object` and this time more of the reads succeed, but yet
          more reads might be generated as well for children of the objects
          which were deserialized.
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self.cache = get_user_cache()
        self.network = get_network()

        # Register for broadcast events
        event_map = {
            KaraboEvent.ProjectItemsLoaded: self._event_items_loaded,
            KaraboEvent.ProjectItemsSaved: self._event_items_saved,
        }
        register_for_broadcasts(event_map)

        # Dictionaries to hold items which are awaiting network replies
        self._waiting_for_read = {}
        self._waiting_for_write = {}
        # Lists of Hashes which are buffered before sending to the GUI server
        self._read_items_buffer = []
        self._write_items_buffer = []

        self.project_manager = KARABO_PROJECT_MANAGER
        self._ignore_cache = True

        # XXX: This is really asinine right now!
        self._have_logged_in = False

    # -------------------------------------------------------------------
    # Karabo Events

    def _event_items_loaded(self, data):
        success = data.get('success')
        items = data.get('items', [])
        self._items_loaded(items, success)

    def _event_items_saved(self, data):
        success = data.get('success')
        items = data.get('items', [])
        self._items_saved(items, success)

    # -------------------------------------------------------------------
    # User interface

    def flush(self):
        """Flush any pending reads or writes.
        """
        if len(self._read_items_buffer) > 0:
            items = self._read_items_buffer
            self.network.onProjectLoadItems(self.project_manager, items)
            self._read_items_buffer = []
        if len(self._write_items_buffer) > 0:
            items = self._write_items_buffer
            self.network.onProjectSaveItems(self.project_manager, items)
            self._write_items_buffer = []

    def get_available_domains(self):
        """ Find out which domains are available
        """
        # Fire and "forget". An event will be broadcast with the reply
        self.network.onListProjectDomains(self.project_manager)
        # Call locally as well
        return self.cache.get_available_domains()

    def get_available_project_data(self, domain, obj_type):
        """ Find out what's available
        """
        if self._ignore_cache:
            # Fire and "forget". An event will be broadcast with the reply
            self.network.onProjectListItems(self.project_manager, domain,
                                            obj_type)
            return []

        # Call locally only if necessary
        return self.cache.get_available_project_data(domain, obj_type)

    def get_projects_with_device(self, domain, device_id):
        """ Find projects which contain configurations for a given device.
        """
        self.network.onProjectListProjectsWithDevice(
            self.project_manager, domain, device_id)

    def get_projects_with_macro(self, domain: str, macro_id: str):
        """ Find projects which contain configurations for a given device.
        """
        self.network.onProjectListProjectsWithMacro(
            self.project_manager, domain, macro_id)

    def get_projects_with_server(self, domain: str, server_id: str):
        """ Find projects which contain configurations for a given server.
        """
        self.network.onProjectListProjectsWithServer(
            self.project_manager, domain, server_id)

    def retrieve(self, domain, uuid, existing=None):
        """Read an object from the database.
        """

        data = None
        if not self._ignore_cache:
            data = self.cache.retrieve(domain, uuid, existing=existing)

        if data is None:
            self._push_reading(domain, uuid, existing)
        return data

    def store(self, domain, uuid, obj):
        """Write an object to the database
        """
        self._push_writing(domain, uuid, obj)

    def update_trashed(self, domain: str, item_type: str, uuid: str,
                       attr_value: bool):
        """ Update any attribute of the of the object
        """
        assert item_type == "project"
        item = Hash('domain', domain, 'uuid', uuid, 'value', attr_value,
                    'item_type', item_type)
        self.network.onProjectUpdateTrashed(self.project_manager, items=item)

    @property
    def default_domain(self):
        return get_config()['domain']

    @default_domain.setter
    def default_domain(self, value):
        get_config()['domain'] = value

    @property
    def ignore_local_cache(self):
        return self._ignore_cache

    @ignore_local_cache.setter
    def ignore_local_cache(self, value):
        self._ignore_cache = value

    def is_reading(self):
        return len(self._waiting_for_read) > 0

    def is_writing(self):
        return len(self._waiting_for_write) > 0

    def is_processing(self):
        return self.is_reading() or self.is_writing()

    def get_database_scene(self, name, uuid, target_window, position=None):
        """Directly ask for a scene from the project database

        This method returns the token for the request.

        Optional keyword arguments:

        - position: The coordinates (List) of the scene
        """
        config = get_config()
        data = {
            "scene_name": name,
            "slot_name": "slotGetScene",
            "target_window": target_window,
            "uuid": uuid,
            "domain": config['domain']}
        if position is not None:
            data["position"] = position

        return get_scene_from_server(self.project_manager, **data)

    # -------------------------------------------------------------------
    # Broadcast event handlers

    def _items_loaded(self, items, success):
        """ A bunch of data just arrived from the network.
        """
        if success:
            # Cache everything locally first only if loading from DB successful
            cache_store = self.cache.store
            for item in items:
                domain = item['domain']
                uuid = item['uuid']
                data = item['xml']
                cache_store(domain, uuid, data)
            storage = self._build_memcache(items)
        else:
            # Project loading failed, use the fast track to tell the world
            self._waiting_for_read.clear()
            self._read_items_buffer = []
            self._broadcast_is_processing(False, bail=True,
                                          loading_failed=True)
            return

        # Then go back through and load any waiting objects
        pop_reading = self._pop_reading
        for item in items:
            domain = item['domain']
            uuid = item['uuid']
            pop_reading(domain, uuid, success, storage)

        # Make a single request to the GUI server
        self.flush()

    def _items_saved(self, items, success):
        """ A bunch of items were just saved
        """
        if success:
            pop_writing = self._pop_writing
            for item in items:
                domain = item['domain']
                uuid = item['uuid']
                date = item['date']
                success = item['success']
                if not success:
                    messagebox.show_error(item['reason'])
                pop_writing(domain, uuid, date, success)
        else:
            # Project writing failed, use the fast track to tell the world
            self._waiting_for_write.clear()
            self._write_items_buffer = []
            self._broadcast_is_processing(False, bail=True)

    # -------------------------------------------------------------------
    # private interface

    def _build_memcache(self, items):
        data = {}
        for item in items:
            domain_data = data.setdefault(item['domain'], {})
            domain_data[item['uuid']] = item['xml']
        return MemCacheWrapper(data, self)

    def _broadcast_is_processing(self, previous_processing, *,
                                 bail=False, loading_failed=False):
        """Create broadcast event and send to all registered ``QObjects``
        """
        if bail:
            # Tell the world reading or writing project failed
            broadcast_event(KaraboEvent.DatabaseIsBusy,
                            {'is_processing': False,
                             'loading_failed': loading_failed})

        # Check current processing state against previous one
        if self.is_processing() == previous_processing:
            return

        # Tell the world
        broadcast_event(KaraboEvent.DatabaseIsBusy,
                        {'is_processing': self.is_processing()})

    def _push_reading(self, domain, uuid, existing):
        # Store previous processing state
        is_processing = self.is_processing()

        if uuid not in self._waiting_for_read:
            assert existing is not None
            self._waiting_for_read[uuid] = existing
            item_type = get_item_type(existing)
            item = Hash('domain', domain, 'uuid', uuid, 'item_type', item_type)
            self._read_items_buffer.append(item)
            if len(self._read_items_buffer) >= MAX_BUFFER_ITEMS:
                self.flush()

        self._broadcast_is_processing(is_processing)

    def _pop_reading(self, domain, uuid, success, storage):
        # Store previous processing state
        is_processing = self.is_processing()

        obj = self._waiting_for_read.pop(uuid, None)
        if obj is not None and success:
            # Only try to read if reading from DB was successful
            read_lazy_object(domain, uuid, storage, read_project_model,
                             existing=obj)

        self._broadcast_is_processing(is_processing)

    def _push_writing(self, domain, uuid, obj):
        # Store previous processing state
        is_processing = self.is_processing()

        # Don't ask the GUI server if you're already waiting for this object
        if uuid not in self._waiting_for_write:
            self._waiting_for_write[uuid] = obj
            # item_type to nofity the clients later
            item_type = get_item_type(obj)
            # Project DB expects xml as string
            xml = write_project_model(obj)
            # XXX overwrite everytime until handled
            item = Hash('domain', domain, 'uuid', uuid,
                        'item_type', item_type, 'xml', xml,
                        'overwrite', True)
            self._write_items_buffer.append(item)
            if len(self._write_items_buffer) >= MAX_BUFFER_ITEMS:
                self.flush()

        self._broadcast_is_processing(is_processing)

    def _pop_writing(self, domain, uuid, date, success):
        # Store previous processing state
        is_processing = self.is_processing()

        obj = self._waiting_for_write.pop(uuid)

        if success:
            # Update date
            obj.date = date
            # Write to the local cache
            data = write_project_model(obj)
            self.cache.store(domain, uuid, data)
            # NOTE: The modified scene model contains one or more widgets
            # that have been modified which need to be reset!
            if isinstance(obj, SceneModel):
                set_modified_flag(obj, False)
            else:
                obj.modified = False

        self._broadcast_is_processing(is_processing)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from unittest import main, mock

from karabogui.binding.api import apply_default_configuration, build_binding
from karabogui.dialogs.api import DeviceCapabilityDialog
from karabogui.testing import (
    GuiTestCase, get_device_schema, singletons, system_hash)
from karabogui.topology.system_topology import SystemTopology


class TestCapabilityDialog(GuiTestCase):

    def test_basic_dialog(self):
        topology = SystemTopology()
        topology.initialize(system_hash())
        network = mock.Mock()
        with singletons(topology=topology, network=network):
            # 1. Test init without deviceId as argument
            dialog = DeviceCapabilityDialog()
            assert not dialog.isModal()
            # We have 1 device in the QListWidget found in topology
            assert dialog.deviceNames.count() == 1

            # Selection of device
            assert dialog.device_id == ""
            dialog.deviceNames.setCurrentRow(0)
            dialog.on_deviceNames_itemSelectionChanged()
            assert dialog.device_id == "divvy"

            # Device in topology with default 'scene`
            device = topology.get_device("divvy")
            build_binding(get_device_schema(), device.binding)
            apply_default_configuration(device.binding)

            # Scene is available and can be selected
            dialog.on_deviceNames_itemSelectionChanged()
            assert dialog.capaNames.count() == 1
            dialog.capaNames.setCurrentRow(0)
            assert dialog.capa_name == "scene"

            # Finish dialog with success!
            dialog.done(1)

            # 2. Test with a deviceId in the beginning!
            dialog = DeviceCapabilityDialog(device_id="divvy")
            assert not dialog.isModal()
            # We have 1 device in the QListWidget found in topology
            assert dialog.deviceNames.count() == 1
            assert dialog.device_id == "divvy"

            # Finish dialog declined!
            dialog.done(0)


if __name__ == "__main__":
    main()



#############################################################################
# Author: <john.wiggins@xfel.eu>
# Created on October 26, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
import json
import re
from contextlib import contextmanager

from qtpy.QtCore import (
    QAbstractItemModel, QItemSelection, QItemSelectionModel, QMimeData,
    QModelIndex, Qt)

from karabo.common.api import walk_traits_object
from karabo.common.project.api import MacroModel
from karabo.common.scenemodel.api import SceneModel
from karabogui.events import KaraboEvent, broadcast_event
from karabogui.indicators import get_instance_info_icon
from karabogui.project.controller.build import (
    create_project_controller, destroy_project_controller)
from karabogui.project.controller.device import DeviceInstanceController
from karabogui.project.utils import show_no_configuration
from karabogui.singletons.api import get_topology
from karabogui.util import create_table_string, utc_to_local

TABLE_HEADER_LABELS = ["Projects", ""]

PROJECT_COLUMN = 0
STATUS_COLUMN = 1

ROLES = [Qt.DisplayRole, Qt.DecorationRole, Qt.ForegroundRole,
         Qt.FontRole, Qt.ToolTipRole]


class ProjectViewItemModel(QAbstractItemModel):
    """ A QAbstractItemModel which mediates between our Traits-based data model
    and Qt
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self.q_selection_model = QItemSelectionModel(self, self)

        self._traits_model = None
        self._controller = None

    def supportedDragActions(self):
        return Qt.CopyAction

    @contextmanager
    def insertion_context(self, parent_controller, first, last):
        """Provide a context for the addition of multiple children under a
        single parent item.

        NOTE: This method is a context manager wraps the insertion with calls
        to ``QAbstractItemModel.beginInsertRows`` and
        ``QAbstractItemModel.endInsertRows`` (See Qt documentation)
        """
        parent_row = self._controller_row(parent_controller)
        parent_index = self.createIndex(parent_row, PROJECT_COLUMN,
                                        parent_controller)
        if parent_index.isValid():
            try:
                self.beginInsertRows(parent_index, first, last)
                yield
            finally:
                self.endInsertRows()

    @contextmanager
    def removal_context(self, controller):
        """Provide a context for the removal of a single item from the model.

        NOTE: This method is a context manager which wraps the removal of an
        item with ``QAbstractItemModel.beginRemoveRows`` and
        ``QAbstractItemModel.endRemoveRows`` (See Qt documentation)
        """
        index_row = self._controller_row(controller)
        index = self.createIndex(index_row, PROJECT_COLUMN, controller)
        if index.isValid():
            parent_index = index.parent()
            if parent_index.isValid():
                try:
                    row = self._controller_row(controller)
                    self.beginRemoveRows(parent_index, row, row)
                    yield
                finally:
                    self.endRemoveRows()

    def controller_data_update(self, controller, column=PROJECT_COLUMN,
                               roles=ROLES):
        index_row = self._controller_row(controller)
        index = self.createIndex(index_row, column, controller)
        if index.isValid():
            self.dataChanged.emit(index, index, roles)

    @property
    def root_controller(self):
        """ Return the project controller object at the root of the hierarchy
        """
        return self._controller

    @property
    def root_model(self):
        """ Return the project object at the root of the hierarchy
        """
        return self._traits_model

    @root_model.setter
    def root_model(self, model):
        """ Set the ProjectModel instance that we're presenting to Qt
        """
        # Clean up any previously created controllers
        if self._controller is not None:
            destroy_project_controller(self._controller)
            get_topology().clear_project_devices()
            self.q_selection_model.clearSelection()
            show_no_configuration()

        self._cleanup_project()

        self._traits_model = model
        try:
            self.beginResetModel()
            if model is not None:
                self._controller = create_project_controller(
                    model=model, parent=None, _qt_model=self)
            else:
                self._controller = None
        finally:
            self.endResetModel()

    def mimeData(self, indices):
        """Reimplemented function of QAbstractItemModel.

        Provide data for Drag & Drop operations.
        """
        # Get one selection per row
        rows = {idx.row(): idx for idx in indices if idx.isValid()}
        # Extract info() dictionaries from controllers
        data = []
        for index in rows.values():
            controller = index.internalPointer()
            if controller is None:
                continue
            data.append(controller.info())

        mimeData = QMimeData()
        mimeData.setData('treeItems', bytearray(json.dumps(data),
                                                encoding='UTF-8'))
        return mimeData

    def selectModel(self, model):
        """Find and select a controller from the project `model`"""
        controller = None

        def _visitor(obj):
            """Find all items matching"""
            if obj.model is model:
                nonlocal controller
                controller = obj
                return True

        self.visit(_visitor)
        if controller is not None:
            self.selectNode(controller)
        return controller is not None

    def findNodes(self, text, case_sensitive=True,
                  use_reg_ex=True, full_match=False):
        """ Find in the ``self._traits_model`` all objects maching criteria"""
        if self._controller is None:
            return []

        controllers = []

        pattern = text if use_reg_ex else f".*{re.escape(text)}"
        flags = 0 if case_sensitive else re.IGNORECASE
        regex = re.compile(pattern, flags=flags)

        matcher = regex.fullmatch if full_match else regex.match

        def _visitor(obj):
            """Find all items matching"""
            if matcher(obj.display_name) is not None:
                controllers.append(obj)

        self.visit(_visitor)
        return controllers

    def visit(self, visitor):
        """Walk every controller in the model and run a `visitor` function on
        each item.
        """

        def _iter_tree_node(node):
            yield node
            for child in getattr(node, 'children', []):
                yield from _iter_tree_node(child)

        for controller in _iter_tree_node(self._controller):
            visitor(controller)

    def selectIndex(self, index):
        """Select the given `index` of type `QModelIndex` if this is not None
        """
        if index is None:
            self.q_selection_model.selectionChanged.emit(
                QItemSelection(), QItemSelection())
            return

        self.q_selection_model.setCurrentIndex(
            index, QItemSelectionModel.ClearAndSelect)

        treeview = super().parent()
        treeview.scrollTo(index)

    def selectNode(self, controller):
        """Select the node matching the given `controller`
        """
        if controller is not None:
            index = self.createIndex(
                self._controller_row(controller),
                PROJECT_COLUMN, controller)
        else:
            # Select nothing
            index = None
        self.selectIndex(index)

    # ----------------------------
    # private methods

    def _cleanup_project(self):
        """ Clean up the ``self._traits_model`` properly which means trigger
        certain events"""
        if self._traits_model is None:
            return

        models = []

        def visitor(obj):
            """Find all macros and scenes"""
            if isinstance(obj, (MacroModel, SceneModel)):
                models.append(obj)

        # Request that views for every macro and scene be closed
        walk_traits_object(self._traits_model, visitor)
        broadcast_event(KaraboEvent.RemoveProjectModelViews,
                        {'models': models})

    def _controller_row(self, controller):
        """Return the row for the given ``controller``

        Note: The row position needs to be fetched from the parent of the given
        ``controller``
        """
        parent_controller = controller.parent
        if parent_controller is None:
            return 0
        else:
            return parent_controller.children.index(controller)

    # ----------------------------
    # Qt methods

    def flags(self, index):
        """Reimplemented function of QAbstractItemModel.
        """
        if not index.isValid():
            return Qt.NoItemFlags

        # All items have these properties
        flags = Qt.ItemIsEnabled | Qt.ItemIsSelectable | Qt.ItemIsDragEnabled
        controller = index.internalPointer()
        if isinstance(controller, DeviceInstanceController):
            flags |= Qt.ItemNeverHasChildren

        return flags

    def data(self, index, role=Qt.DisplayRole):
        """Reimplemented function of QAbstractItemModel.
        """
        if not index.isValid():
            return

        controller = index.internalPointer()
        if controller is None:
            return

        column = index.column()
        ui_data = controller.ui_data
        if column == PROJECT_COLUMN:
            if role == Qt.DisplayRole:
                return controller.ui_item_text()
            elif (role == Qt.ToolTipRole and not isinstance(
                    controller, DeviceInstanceController)):
                return self._create_tool_tip(controller.model)
            elif role == Qt.DecorationRole:
                return ui_data.icon
            elif role == Qt.ForegroundRole:
                return ui_data.brush
            elif role == Qt.FontRole:
                return ui_data.font
        elif column == STATUS_COLUMN and role == Qt.DecorationRole:
            if isinstance(controller, DeviceInstanceController):
                return get_instance_info_icon(ui_data.instance_status)

    def _create_tool_tip(self, model):
        info = {"uuid": model.uuid,
                "modified": utc_to_local(model.date)}

        return create_table_string(info)

    def index(self, row, column, parent=QModelIndex()):
        """Reimplemented function of QAbstractItemModel.
        """
        if self._controller is None:
            return QModelIndex()

        if not self.hasIndex(row, column, parent):
            return QModelIndex()

        if not parent.isValid():
            return self.createIndex(row, column, self._controller)
        else:
            parent_controller = parent.internalPointer()

        if parent_controller is None:
            return QModelIndex()

        child_controller = parent_controller.child(row)
        if child_controller is not None:
            return self.createIndex(row, column, child_controller)

        return QModelIndex()

    def parent(self, index):
        """Reimplemented function of QAbstractItemModel.
        """
        if self._controller is None:
            return QModelIndex()

        if not index.isValid():
            return QModelIndex()

        child_controller = index.internalPointer()
        if child_controller is None:
            return QModelIndex()

        parent_controller = child_controller.parent
        if parent_controller is None:
            return QModelIndex()

        return self.createIndex(self._controller_row(parent_controller),
                                PROJECT_COLUMN, parent_controller)

    def headerData(self, section, orientation, role):
        """Reimplemented function of QAbstractItemModel.
        """
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return TABLE_HEADER_LABELS[section]

    def columnCount(self, parentIndex=QModelIndex()):
        """Reimplemented function of QAbstractItemModel.
        """
        return len(TABLE_HEADER_LABELS)

    def rowCount(self, parent=QModelIndex()):
        """Reimplemented function of QAbstractItemModel.
        """
        # Only column 0 has children
        if self._controller is None or parent.column() > 0:
            return 0

        if not parent.isValid():
            return 1
        else:
            parent_controller = parent.internalPointer()

        if parent_controller is None:
            return 0

        return parent_controller.rows()



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# Created on July 9, 2019
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QAction, QFrame, QHBoxLayout, QInputDialog, QLabel, QSlider, QWidget)
from traits.api import Bool, Instance, on_trait_change

from karabo.common.api import (
    KARABO_SCHEMA_MAX_EXC, KARABO_SCHEMA_MAX_INC, KARABO_SCHEMA_MIN_EXC,
    KARABO_SCHEMA_MIN_INC)
from karabo.common.scenemodel.api import TickSliderModel
from karabogui.binding.api import (
    FloatBinding, IntBinding, get_editor_value, get_min_max,
    has_min_max_attributes)
from karabogui.controllers.api import (
    BaseBindingController, is_proxy_allowed, register_binding_controller)
from karabogui.util import SignalBlocker

# Define a maximum range which can be represented by ticks
REASONABLE_RANGE = 100000
TICK_RANGE = 1000
MIN_TICK = 1
WIDGET_HEIGHT = 20
WIDGET_WIDTH = 60


@register_binding_controller(ui_name='Tick Slider', can_edit=True,
                             klassname='TickSlider',
                             is_compatible=has_min_max_attributes,
                             binding_type=(FloatBinding, IntBinding))
class TickSlider(BaseBindingController):
    # The scene model class for this controller
    model = Instance(TickSliderModel, args=())

    slider = Instance(QSlider)
    label = Instance(QLabel)
    _has_error = Bool(False)

    def create_widget(self, parent):
        widget = QWidget(parent)
        layout = QHBoxLayout(widget)
        widget.setLayout(layout)

        self.slider = QSlider(Qt.Horizontal, widget)
        self.slider.setTickPosition(QSlider.TicksBelow)
        self.slider.setFocusPolicy(Qt.StrongFocus)
        self.slider.valueChanged.connect(self._edit_value)
        self.slider.setTickInterval(self.model.ticks)
        self.slider.setSingleStep(self.model.ticks)

        layout.addWidget(self.slider)

        self.label = QLabel(widget)
        self.label.setFrameStyle(QFrame.Box)
        self.label.setAlignment(Qt.AlignCenter)
        self.label.setVisible(self.model.show_value)
        self.label.setFixedSize(WIDGET_WIDTH, WIDGET_HEIGHT)
        self.label.setStyleSheet("font: 6pt; font-weight: bold;")
        layout.addWidget(self.label)

        action_ticks = QAction("Tick Interval", widget)
        action_ticks.triggered.connect(self.configure_tick_interval)
        widget.addAction(action_ticks)

        action_show_value = QAction("Show value", widget, checkable=True)
        action_show_value.toggled.connect(self.toggle_show_value)
        action_show_value.setChecked(self.model.show_value)
        widget.addAction(action_show_value)

        return widget

    def binding_update(self, proxy):
        attrs = proxy.binding.attributes
        min_inc = attrs.get(KARABO_SCHEMA_MIN_INC)
        min_exc = attrs.get(KARABO_SCHEMA_MIN_EXC)
        max_inc = attrs.get(KARABO_SCHEMA_MAX_INC)
        max_exc = attrs.get(KARABO_SCHEMA_MAX_EXC)
        low, high = get_min_max(proxy.binding)
        # convert to native int, subtracting limits can overflow
        # on numpy subtractions
        low = int(low)
        high = int(high)
        out_of_range = high - low
        out_of_range = out_of_range > REASONABLE_RANGE
        if (min_inc is None and min_exc is None or
                max_inc is None and max_exc is None or
                out_of_range):
            with SignalBlocker(self.widget):
                self._error(proxy.path)
            return

        with SignalBlocker(self.slider):
            if high - low > TICK_RANGE:
                self.slider.setTickPosition(QSlider.NoTicks)
            self.slider.setRange(low, high)

    def value_update(self, proxy):
        value = get_editor_value(proxy)
        if value is not None:
            with SignalBlocker(self.slider):
                self.slider.setValue(int(value))
                if isinstance(proxy.binding, IntBinding):
                    fmt = "{}"
                elif isinstance(proxy.binding, FloatBinding):
                    fmt = "{:.1f}"
                self.label.setText(fmt.format(value))

    def _error(self, keyname):
        """Configure the widget as a disabled one

        call this function inside a signal blocker context"""
        msg = ('The value limits for {} are not set or too large (>{}),'
               ' please check the property attributes'.format(
                   keyname, REASONABLE_RANGE))
        self._has_error = True
        self.widget.setToolTip(msg)
        self.widget.setEnabled(False)

    def state_update(self, proxy):
        if self._has_error:
            # In case we are disabled, we do not consider state updates!
            return

        enable = is_proxy_allowed(proxy)
        self.widget.setEnabled(enable)

    @on_trait_change('model:show_value')
    def _show_value_update(self, value):
        self.label.setVisible(value)

    # ----------------------------------------------------------------------
    # Qt Slots

    def _edit_value(self, value):
        if self.proxy.binding is None:
            return
        self.proxy.edit_value = value
        if value is not None:
            if isinstance(self.proxy.binding, IntBinding):
                fmt = "{}"
            elif isinstance(self.proxy.binding, FloatBinding):
                fmt = "{:.1f}"
            self.label.setText(fmt.format(value))

    def toggle_show_value(self, value):
        self.model.show_value = value

    def configure_tick_interval(self):
        """Configure the tick interval for this widget"""
        ticks, ok = QInputDialog.getInt(self.widget, 'Tick Interval',
                                        'Tick Interval:', self.model.ticks,
                                        MIN_TICK, TICK_RANGE)
        if ok:
            self.model.ticks = ticks
            self.slider.setTickInterval(ticks)
            self.slider.setSingleStep(ticks)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import pytest
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QDialog, QToolButton

from karabo.common.scenemodel.api import (
    EditableListModel, EditableRegexListModel)
from karabo.native import (
    Configurable, VectorInt8, VectorInt32, VectorRegexString)
from karabogui.binding.api import apply_default_configuration, get_min_max_size
from karabogui.dialogs.listedit import ListEditDialog
from karabogui.testing import (
    click_button, get_class_property_proxy, set_proxy_value)

from ..list import EditableList, EditableRegexList


class Object(Configurable):
    prop = VectorInt32(defaultValue=[1])


class SizeObject(Configurable):
    prop = VectorInt32(defaultValue=[1], minSize=1, maxSize=3)


class ShortObject(Configurable):
    prop = VectorInt8(defaultValue=[1])


class RegexObject(Configurable):
    prop = VectorRegexString(
        defaultValue=["remote:output"],
        regex=r"^[A-Za-z0-9_-]{1,60}(:)[A-Za-z0-9_-]{1,60}$")


class ListEditMock(ListEditDialog):
    @property
    def values(self):
        return ["-1", "42", "-1"]

    def exec(self):
        return QDialog.Accepted


@pytest.fixture
def list_setup(gui_app):
    proxy = get_class_property_proxy(Object.getClassSchema(), "prop")
    controller = EditableList(proxy=proxy, model=EditableListModel())
    controller.create(None)
    apply_default_configuration(proxy.root_proxy.binding)
    yield controller, proxy
    # teardown
    controller.destroy()
    assert controller.widget is None


def test_list_focus_policy(list_setup):
    controller, _ = list_setup
    assert controller._internal_widget.focusPolicy() == Qt.StrongFocus


def test_list_set_value(list_setup):
    controller, proxy = list_setup
    controller.last_cursor_position = 0
    set_proxy_value(proxy, "prop", [0, 2])
    assert controller._internal_widget.text() == "0,2"
    assert controller._internal_widget.cursorPosition() == 0


def test_list_edit_value(list_setup):
    controller, proxy = list_setup
    controller.set_read_only(False)
    controller._internal_widget.setText("3,4")
    assert all(proxy.edit_value == [3, 4])


def test_list_edit_empty_value(list_setup):
    controller, proxy = list_setup
    controller._internal_widget.setText("NOT IMPORTANT")
    controller.set_read_only(False)
    controller._internal_widget.setText("")
    value = proxy.edit_value
    assert len(value) == 0 and all(value == [])


def test_list_edit_dialog(list_setup, mocker):
    controller, proxy = list_setup
    target = "karabogui.controllers.edit.list.ListEditDialog"
    mocker.patch(target, new=ListEditMock)
    controller.set_read_only(False)
    controller._on_edit_clicked()
    assert all(proxy.edit_value == [-1, 42, -1])


@pytest.fixture
def list_size_setup(gui_app):
    size_proxy = get_class_property_proxy(SizeObject.getClassSchema(), "prop")
    size_controller = EditableList(proxy=size_proxy,
                                   model=EditableListModel())
    size_controller.create(None)
    apply_default_configuration(size_proxy.root_proxy.binding)
    yield size_controller, size_proxy
    # teardown
    size_controller.destroy()
    assert size_controller.widget is None


def test_size_list(list_size_setup):
    size_controller, size_proxy = list_size_setup
    assert get_min_max_size(size_proxy.binding) == (1, 3)
    size_controller.set_read_only(False)
    size_controller._internal_widget.setText("3,4")
    assert all(size_proxy.edit_value == [3, 4])
    size_controller._internal_widget.setText("3,4,5,6")
    assert size_proxy.edit_value is None
    size_controller._internal_widget.setText("")
    assert size_proxy.edit_value is None


def test_size_list_trailing_zeros(list_size_setup):
    size_controller, size_proxy = list_size_setup
    assert get_min_max_size(size_proxy.binding) == (1, 3)
    size_controller.set_read_only(False)
    size_controller._internal_widget.setText("03,4")
    assert size_proxy.edit_value is None
    size_controller._internal_widget.setText("3,4")
    assert all(size_proxy.edit_value == [3, 4])
    size_controller._internal_widget.setText("000000,4")
    assert all(size_proxy.edit_value == [0, 4])


def test_list_shorts(gui_app):
    # setup
    short_proxy = get_class_property_proxy(ShortObject.getClassSchema(),
                                           "prop")
    short_controller = EditableList(proxy=short_proxy,
                                    model=EditableListModel())
    short_controller.create(None)
    apply_default_configuration(short_proxy.root_proxy.binding)
    # test body
    short_controller.set_read_only(False)
    short_controller._internal_widget.setText("03,4")
    assert short_proxy.edit_value is None
    short_controller._internal_widget.setText("3,4")
    assert all(short_proxy.edit_value == [3, 4])
    # out of range
    short_controller._internal_widget.setText("10300")
    assert short_proxy.edit_value is None
    # teardown
    short_controller.destroy()
    assert short_controller.widget is None


@pytest.fixture
def editable_regex_list_setup(gui_app):
    proxy = get_class_property_proxy(RegexObject.getClassSchema(), "prop")
    controller = EditableRegexList(proxy=proxy, model=EditableRegexListModel())
    controller.create(None)
    controller.binding_update(proxy)
    apply_default_configuration(proxy.root_proxy.binding)
    yield controller, proxy
    # teardown
    controller.destroy()
    assert controller.widget is None


def test_editable_regex_list_binding_update_validator(
        editable_regex_list_setup):
    controller, _ = editable_regex_list_setup
    assert controller._internal_widget.validator() is not None


def test_editable_regex_list_set_value(editable_regex_list_setup):
    controller, proxy = editable_regex_list_setup
    controller.last_cursor_position = 0
    set_proxy_value(proxy, "prop", ["remote:outputDaq"])
    assert controller._internal_widget.text() == "remote:outputDaq"
    assert controller._internal_widget.cursorPosition() == 0


def test_editable_regex_list_edit_value(editable_regex_list_setup):
    controller, proxy = editable_regex_list_setup
    controller.set_read_only(False)
    controller._internal_widget.setText("remote:outputDaq,remote:output")
    assert proxy.edit_value == ["remote:outputDaq", "remote:output"]


def test_editable_regex_list_edit_empty_value(editable_regex_list_setup):
    controller, proxy = editable_regex_list_setup
    controller.set_read_only(False)
    controller._internal_widget.setText("remote:output")
    assert proxy.edit_value == ["remote:output"]
    controller._internal_widget.setText("")
    value = proxy.edit_value
    assert value == []


def test_editable_list_dialog_value(editable_regex_list_setup, mocker):
    controller, proxy = editable_regex_list_setup
    controller.last_cursor_position = 0
    set_proxy_value(proxy, "prop", ["1", "2", "3"])
    button = controller.widget.findChild(QToolButton)
    path = "karabogui.controllers.edit.list.ListEditDialog"
    dialog = mocker.patch(path)
    dialog().exec.return_value = QDialog.Accepted
    dialog().values = ["1", "2", "3", "4"]
    click_button(button)
    assert controller._internal_widget.text() == "1,2,3,4"
    assert controller._internal_widget.cursorPosition() == 0



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
# flake8: noqa
from .configurationpanel import ConfigurationPanel
from .container import PanelContainer
from .devicepanel import DevicePanel
from .macropanel import MacroPanel
from .navigationpanel import TopologyPanel
from .placeholderpanel import PlaceholderPanel
from .projectpanel import ProjectPanel
from .scenepanel import ScenePanel
from .widgetcontrollerpanel import WidgetControllerPanel



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
CELL_WIDTH = 50


def table_header(header="&nbsp;", tabs=tuple()):
    """Composes an html header row"""
    header_html = f"<th align='left' width='{CELL_WIDTH}'>{header}</th>"
    tabs_html = "".join([f"<th align='left' width='{CELL_WIDTH}'>{tab}</th>"
                         for tab in tabs])
    return header_html + tabs_html


def table_body(header="&nbsp;", tabs=tuple()):
    """Composes an html table row"""
    header_html = f"<th align='left' width='{CELL_WIDTH}'>{header}</th>"
    tabs_html = "".join([f"<td width='{CELL_WIDTH}'>{tab}</td>"
                         for tab in tabs])
    return header_html + tabs_html


def table_row(*rows):
    rows = f"<td width='{CELL_WIDTH / 3}'></td>".join(rows)
    return f"<tr>{rows}</tr>"



from pathlib import Path

from qtpy import uic
from qtpy.QtCore import QObject, Qt, Slot
from qtpy.QtGui import QColor, QIcon, QPixmap
from qtpy.QtWidgets import (
    QDialog, QDialogButtonBox, QListWidgetItem, QMessageBox)

from karabo.native import is_equal
from karabogui.graph.common.api import get_available_colors, rgba_to_hex


class CurveOptionsDialog(QDialog):

    def __init__(self, curve_options: dict | None = None,
                 parent: QObject = None):
        super().__init__(parent=parent)
        ui_file = Path(__file__).parent / "curve_options.ui"
        uic.loadUi(ui_file, self)
        self.setModal(False)
        self.has_reset = False
        self._pen_color = QColor()
        self._prepare_color_combobox()
        self._original = curve_options
        self._load_options(curve_options)
        self.proxy_list.setCurrentRow(0)

        self.legend_name.editingFinished.connect(self._on_legend_name_changed)
        self.color_combo_box.currentIndexChanged.connect(
            self.on_color_changed)

        self.buttonBox.button(QDialogButtonBox.Ok).setText("Apply")
        self.buttonBox.button(QDialogButtonBox.Cancel).setText("Discard")
        self.buttonBox.button(QDialogButtonBox.Reset).clicked.connect(
            self._set_reset)
        # XXX: not fully implemented yet
        self.symbol_group_box.setVisible(False)

    def _prepare_color_combobox(self):
        """Load the combobox with the  allowed colors for the curve"""

        # Fixed width for the icon as the text is empty.
        self.color_combo_box.setFixedWidth(50)
        for single_letter, rgba_code in get_available_colors():
            pixmap = QPixmap(32, 32)
            pixmap.fill(QColor(*rgba_code))
            icon = QIcon(pixmap)
            hexa_code = rgba_to_hex(rgba_code)
            self.color_combo_box.addItem(icon, "", hexa_code)

    def _load_options(self, curve_options: dict):
        """Create a list item for each curve and store their options as
        user data."""
        for curve, data in curve_options.items():
            key = data["key"]
            list_item = QListWidgetItem(key)
            list_item.setData(Qt.UserRole, data)
            list_item.setData(Qt.UserRole + 1, curve)
            self.proxy_list.addItem(list_item)

    @Slot(int)
    def on_proxy_list_currentRowChanged(self, row):
        """Display the curve options for the selected curve"""
        item = self.proxy_list.item(row)
        if item is None:
            self.settings_widget.setEnabled(False)
            return
        self.settings_widget.setEnabled(True)
        curve_options = item.data(Qt.UserRole)
        self.property_key.setText(curve_options["key"])
        self.legend_name.setText(curve_options["name"])
        pen_color = curve_options.get("pen_color")
        index = self.color_combo_box.findData(pen_color)
        if index == -1:
            self.line_group_box.setEnabled(False)
        else:
            self._pen_color = QColor(pen_color)
            self.color_combo_box.setCurrentIndex(index)

    @Slot()
    def _on_legend_name_changed(self):
        """Save the legend name for the selected curve"""
        legend_name = self.legend_name.text()
        self._update_item_data("name", legend_name)

    @Slot(int)
    def on_color_changed(self, index: int):
        """Save the pen color for the selected curve"""
        pen_color = self.color_combo_box.itemData(index)
        self._update_item_data("pen_color", pen_color)

    @Slot()
    def _set_reset(self):
        """Emit a signal to discard all the custom curve options on
        accepting the question dialog."""
        reply = QMessageBox.question(
            self, "Reset All",
            "Do you want to reset to the default options?",
            QMessageBox.Yes | QMessageBox.No)
        if reply == QMessageBox.No:
            return
        self.has_reset = True
        self.accept()

    def _update_item_data(self, key: str, value: str):
        """A convenient method to set the data of the current item in the
        curve option"""
        item = self.proxy_list.currentItem()
        if item is None:
            return
        data = item.data(Qt.UserRole)
        data[key] = value
        item.setData(Qt.UserRole, data)

    def get_curve_options(self) -> dict | None:
        """Return the plotting options for each plot"""
        if self.has_reset:
            return None

        curve_options = {}
        for index in range(self.proxy_list.count()):
            item = self.proxy_list.item(index)
            data = item.data(Qt.UserRole)
            curve = item.data(Qt.UserRole + 1)
            curve_options[curve] = data

        changes = not is_equal(curve_options, self._original)
        return curve_options if changes else {}



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from pytestqt.modeltest import ModelTester
from qtpy.QtCore import QItemSelectionModel, QPoint, Qt

from karabogui.dialogs.api import ClientTopologyDialog
from karabogui.testing import singletons, system_hash
from karabogui.topology.system_topology import SystemTopology


def test_basic_dialog(gui_app, mocker):
    topology = SystemTopology()
    topology.initialize(system_hash())
    manager = mocker.Mock()
    with singletons(topology=topology, manager=manager):
        dialog = ClientTopologyDialog()
        assert not dialog.isModal()
        model = dialog.tree_view.model()
        assert model is not None
        # We have 1 client
        assert model.rowCount() == 1
        client_index = model.index(0, 0)
        assert client_index.data() == "charlie"

        assert model.headerData(
            0, Qt.Horizontal, Qt.DisplayRole) == 'Client Id'
        assert model.headerData(1, Qt.Horizontal, Qt.DisplayRole) == 'Host'

        flag = model.flags(client_index)
        assert flag & Qt.ItemIsEnabled == Qt.ItemIsEnabled
        assert flag & Qt.ItemIsSelectable == Qt.ItemIsSelectable
        assert flag & Qt.ItemIsEditable != Qt.ItemIsEditable

        # Check context menu without selection
        pos = QPoint(0, 0)
        menu = mocker.patch("karabogui.dialogs.client_topology.QMenu")
        dialog._context_menu(pos)
        menu.assert_not_called()

        # Select the client and kill!
        selection = dialog.tree_view.selectionModel()
        selection.setCurrentIndex(client_index,
                                  QItemSelectionModel.ClearAndSelect)
        dialog.onKillInstance()
        manager.shutdownDevice.assert_called_with("charlie",
                                                  parent=dialog)

        # Check context menu with selection!
        pos = QPoint(0, 0)
        dialog._context_menu(pos)
        menu.assert_called_once()


def test_model_tester():
    topology = SystemTopology()
    topology.initialize(system_hash())
    with singletons(topology=topology):
        dialog = ClientTopologyDialog()
        model = dialog.tree_view.model()
        tester = ModelTester(None)
        tester.check(model)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np
from pyqtgraph import PlotItem

from karabogui.graph.plots.base import KaraboPlotView
from karabogui.graph.plots.tools import CrossTargetController

X_ARRAY = np.arange(10)
Y_ARRAY = X_ARRAY ** 2


def test_logMode(gui_app):
    plot_item = PlotItem()
    controller = CrossTargetController(plot_item)
    assert controller._get_x_value(0.0457) == "0.046"
    assert controller._get_y_value(1) == "1"
    assert controller._get_y_value(2) == "2"
    assert controller._get_y_value(2.823) == "2.823"

    plot_item.setLogMode(x=True, y=True)
    controller = CrossTargetController(plot_item)

    # Value should be in 'e' notation
    assert controller._get_x_value(0.0457) == "1.111e+00"
    assert controller._get_y_value(1) == "1.000e+01"
    assert controller._get_y_value(2) == "1.000e+02"
    assert controller._get_y_value(2.823) == "6.653e+02"


def test_legend_color(gui_app):
    """Test the Target tool legend color is black"""
    widget = KaraboPlotView()
    plot = widget.add_curve_item()
    plot.setData(X_ARRAY, Y_ARRAY)
    plot_item = widget.plotItem

    target_tool = CrossTargetController(plot_item)
    target_tool.activate()
    legend = target_tool.legend
    legend_color = legend._label.opts.get("color")
    assert legend_color == "k"
